"""
Extension hook system based on Flask's exthook.py, originally written by Armin
Ronacher and licensed under BSD license.
"""


import sys


class ExtensionImporter(object):

    def __init__(self, wrapper_module):
        self.wrapper_module = wrapper_module
        self.prefix = wrapper_module + '.'
        self.prefix_cutoff = wrapper_module.count('.') + 1

    def __eq__(self, other):
        return self.__class__.__module__ == other.__class__.__module__ and \
            self.__class__.__name__ == other.__class__.__name__ and \
            self.wrapper_module == other.wrapper_module

    def __ne__(self, other):
        return not self.__eq__(other)

    def install(self):
        sys.meta_path[:] = [x for x in sys.meta_path if self != x] + [self]

    def find_module(self, fullname, path=None):
        if fullname.startswith(self.prefix):
            return self

    def load_module(self, fullname):
        if fullname in sys.modules:
            return sys.modules[fullname]
        modname = fullname.split('.', self.prefix_cutoff)[self.prefix_cutoff]
        realname = 'concert_{}'.format(modname)
        __import__(realname)
        module = sys.modules[fullname] = sys.modules[realname]

        if '.' not in modname:
            setattr(sys.modules[self.wrapper_module], modname, module)

        return module
        raise ImportError('No module named %s' % fullname)



import numpy as np
from concert.quantities import q
from concert.devices.cameras.dummy import Base as DummyBaseCamera


def get_np_angle(angle):
    return angle.to(q.rad).magnitude


def rot_x(angle, matrix):
    """Rotation around x-axis."""
    angle = get_np_angle(angle)
    m_0 = np.array([[1, 0, 0],
                    [0, np.cos(angle), -np.sin(angle)],
                    [0, np.sin(angle), np.cos(angle)]])
    new_mat = np.identity(4, np.float)
    new_mat[:3, :3] = m_0

    return np.dot(new_mat, matrix)


def rot_y(angle, matrix):
    """Rotation around y-axis."""
    angle = get_np_angle(angle)
    m_0 = np.array([[np.cos(angle), 0, np.sin(angle)],
                    [0, 1, 0],
                    [-np.sin(angle), 0, np.cos(angle)]])
    new_mat = np.identity(4, np.float)
    new_mat[:3, :3] = m_0

    return np.dot(new_mat, matrix)


def rot_z(angle, matrix):
    """Rotation around z-axis."""
    angle = get_np_angle(angle)
    m_0 = np.array([[np.cos(angle), -np.sin(angle), 0],
                    [np.sin(angle), np.cos(angle), 0],
                    [0, 0, 1]])
    new_mat = np.identity(4, np.float)
    new_mat[:3, :3] = m_0

    return np.dot(new_mat, matrix)


def scale(scales, matrix):
    """Scaling amongst all directions."""
    scales = scales + (1,)

    return np.dot(np.diag(scales), matrix)


def translate(vec, matrix):
    """3D translation."""
    matrix[0, -1] += vec[0]
    matrix[1, -1] += vec[1]
    matrix[2, -1] += vec[2]

    return matrix


def sphere(size, radius, mat):
    """Create a sphere with radius *radius* in an image with size *size* and
    use transformation matrix *mat* to adjust the sphere position and size.
    """
    y_0, x_0 = np.mgrid[-size // 2:size // 2, -size // 2:size // 2][:, ::-1]

    k_x = x_0 * mat[0, 0] + y_0 * mat[0, 1] + mat[0, 3]
    k_y = x_0 * mat[1, 0] + y_0 * mat[1, 1] + mat[1, 3]
    k_z = x_0 * mat[2, 0] + y_0 * mat[2, 1] + mat[2, 3]

    quadr_coeffs = mat[0, 2] ** 2 + mat[1, 2] ** 2 + mat[2, 2] ** 2, \
        2 * k_x * mat[0, 2] + 2 * k_y * mat[1, 2] + 2 * k_z * mat[2, 2], \
        k_x ** 2 + k_y ** 2 + k_z ** 2 - radius ** 2

    thickness = quadr_coeffs[1] ** 2 - 4 * quadr_coeffs[0] * quadr_coeffs[2]
    thickness[thickness < 0] = 0

    return np.abs(np.sqrt(thickness) / quadr_coeffs[0])


def transfer(thickness, ref_index, lam):
    lam = lam.to(thickness.units)
    mju = 4 * q.dimensionless * np.pi * ref_index.imag / lam

    return (mju * thickness).to_base_units().magnitude


def get_projection(thickness):
    """Get X-ray projection image from projected thickness."""
    max_val = thickness.max()
    if max_val > 0:
        thickness = thickness / thickness.max()

    # Iron and 20 keV
    ref_index = 3.85263274e-06 + 9.68238822e-08j
    lam = 6.1992e-11 * q.m

    # Do not take noise into account in order to make test results
    # reproducible.
    return transfer(thickness * q.mm, ref_index, lam)


class SimulationCamera(DummyBaseCamera):

    """A dummy image source providing images of a rotated sample. Rotation
    is based on virtual motors.
    """
    ITER = "iteration"

    def __init__(self, im_width, x_param, y_param, z_param,
                 needle_radius=None, rotation_radius=None,
                 y_position=None, scales=None):
        self.x_axis_param = x_param
        self.y_axis_param = y_param
        self.z_axis_param = z_param
        self.size = im_width
        self._center = None
        self._ellipsoid_center = None
        self.radius = needle_radius
        if self.radius is None:
            self.radius = self.size / 8

        self.rotation_radius = rotation_radius if rotation_radius is not None\
            else self.size / 6
        self.y_position = y_position if y_position \
            is not None else self.size / 2
        self.scale = scales if scales is not None else (3, 0.75, 3)

        # How many times was the image source asked for images.
        self.iteration = 0

        super(SimulationCamera, self).__init__()

    @property
    def ellipse_center(self):
        """Ellipse center."""
        return self._center

    @property
    def ellipsoid_center(self):
        """Ellipsoid center."""
        return self._ellipsoid_center

    async def create_needle(self):
        """Create sample rotated about axis of rotation."""
        matrix = np.identity(4, np.float)

        matrix = translate((0, self.y_position, 0), matrix)

        matrix = rot_z(await self.z_axis_param.get(), matrix)
        matrix = rot_x(await self.x_axis_param.get(), matrix)

        center = np.dot(np.linalg.inv(matrix),
                        (0, self.size / 8 / self.scale[1], 0, 1)) + self.size / 2
        # Ellipse center.
        self._center = self.size - center[1], center[0]

        matrix = rot_y(await self.y_axis_param.get(), matrix)
        matrix = translate((self.rotation_radius, 0, 0), matrix)

        matrix = scale(self.scale, matrix)
        x, y = np.dot(np.linalg.inv(matrix), (0, 0, 0, 1))[:-2] + self.size / 2
        self._ellipsoid_center = (x, self.size - y - 1)

        return sphere(self.size, self.radius, matrix)

    async def _grab_real(self):
        return get_projection(await self.create_needle())



"""Dummy lightsource implementation"""

from concert.quantities import q
from concert.devices.lightsources import base


class LightSource(base.LightSource):

    """A dummy light source"""

    def __init__(self):
        super(LightSource, self).__init__()
        self._intensity = 0 * q.V

    async def _set_intensity(self, intensity):
        self._intensity = intensity

    async def _get_intensity(self):
        return self._intensity



"""
A :class:`Camera` can be set via the device-specific properties that can be set
and read with :meth:`.Parameter.set` and :meth:`.Parameter.get`.  Moreover, a
camera provides means to

* :meth:`~Camera.start_recording` frames,
* :meth:`~Camera.stop_recording` the acquisition,
* :meth:`~Camera.trigger` a frame capture and
* :meth:`~Camera.grab` to get the last frame.

Camera triggering is specified by the :attr:`~Camera.trigger_source` parameter, which
can be one of

* :attr:`camera.trigger_sources.AUTO` means the camera triggers itself
  automatically, the frames start being recorded right after the
  :meth:`~camera.start_recording` call and stop being recorded by
  :meth:`~camera.stop_recording`

* :attr:`Camera.trigger_sources.SOFTWARE` means the camera needs to be triggered
  by the user by :meth:`~Camera.trigger()`. This way you have complete programatic
  control over when is the camera triggered, example usage::

    camera.trigger_source = camera.trigger_sources.SOFTWARE
    start_recording(camera)
    trigger(camera)
    long_operation()
    # Here we get the frame from before the long operation
    grab(camera)

* :attr:`Camera.trigger_sources.EXTERNAL` is a source when the camera is triggered
  by an external low-level signal (such as TTL). This source provides very precise
  triggering in terms of time synchronization with other devices


To setup and use a camera in a typical environment, you would do::

    import numpy as np
    from concert.devices.cameras.uca import Camera

    camera = Camera('pco')
    camera.trigger_source = camera.trigger_sources.SOFTWARE
    camera.exposure_time = 0.2 * q.s
    start_recording(camera)
    trigger(camera)
    data = grab(camera)
    stop_recording(camera)

    print("mean=%f, stddev=%f" % (np.mean(data), np.std(data)))

You can apply primitive operations to the frames obtained by :meth:`Camera.grab` by setting up a
:attr:`Camera.convert` attribute to some callable which takes just one argument which is the grabbed
frame. The callable is applied to the frame and the converted one is returned by
:meth:`Camera.grab`. You can do::

    import numpy as np
    from concert.devices.cameras.dummy import Camera

    camera = Camera()
    camera.convert = np.fliplr
    # The frame is left-right flipped
    grab(camera)
"""
import contextlib
import logging
from concert.base import AccessorNotImplementedError, Parameter, Quantity, State, check, identity
from concert.config import AIODEBUG
from concert.coroutines.base import background
from concert.quantities import q
from concert.helpers import Bunch
from concert.devices.base import Device


LOG = logging.getLogger(__name__)


class CameraError(Exception):

    """Camera specific errors."""
    pass


class Camera(Device):

    """Base class for remotely controllable cameras.

    .. py:attribute:: frame-rate

        Frame rate of acquisition in q.count per time unit.
    """

    trigger_sources = Bunch(['AUTO', 'SOFTWARE', 'EXTERNAL'])
    trigger_types = Bunch(['EDGE', 'LEVEL'])
    state = State(default='standby')
    frame_rate = Quantity(1 / q.second, help="Frame frequency")
    trigger_source = Parameter(help="Trigger source")

    def __init__(self):
        super(Camera, self).__init__()
        self.convert = identity

    @background
    @check(source='standby', target='recording')
    async def start_recording(self):
        """
        start_recording()

        Start recording frames.
        """
        await self._record_real()

    @background
    @check(source='recording', target='standby')
    async def stop_recording(self):
        """
        stop_recording()

        Stop recording frames.
        """
        await self._stop_real()

    @contextlib.asynccontextmanager
    async def recording(self):
        """
        recording()

        A context manager for starting and stopping the camera.

        In general it is used with the ``async with`` keyword like this::

            async with camera.recording():
                frame = await camera.grab()
        """
        await self.start_recording()
        try:
            yield
        finally:
            LOG.log(AIODEBUG, 'stop recording in recording()')
            await self.stop_recording()

    @background
    async def trigger(self):
        """Trigger a frame if possible."""
        await self._trigger_real()

    @background
    async def grab(self):
        """Return a NumPy array with data of the current frame."""
        return self.convert(await self._grab_real())

    async def stream(self):
        """
        stream()

        Grab frames continuously yield them. This is an async generator.
        """
        await self.set_trigger_source(self.trigger_sources.AUTO)
        await self.start_recording()

        while await self.get_state() == 'recording':
            yield await self.grab()

    async def _get_trigger_source(self):
        raise AccessorNotImplementedError

    async def _set_trigger_source(self, source):
        raise AccessorNotImplementedError

    async def _record_real(self):
        raise AccessorNotImplementedError

    async def _stop_real(self):
        raise AccessorNotImplementedError

    async def _trigger_real(self):
        raise AccessorNotImplementedError

    async def _grab_real(self):
        raise AccessorNotImplementedError


class BufferedMixin(Device):

    """A camera that stores the frames in an internal buffer"""

    state = State(default='standby')

    def readout_buffer(self, *args, **kwargs):
        return self._readout_real(*args, **kwargs)

    async def _readout_real(self, *args, **kwargs):
        raise AccessorNotImplementedError



"""
An X-ray tube.
"""
from concert.base import check
from concert.coroutines.base import background
from concert.quantities import q
from concert.base import Quantity, AccessorNotImplementedError, State
from concert.devices.base import Device


class XRayTube(Device):
    """
    A base x-ray tube class.
    """

    voltage = Quantity(q.kV)
    current = Quantity(q.uA)
    power = Quantity(q.W)

    state = State(default='off')

    def __init__(self):
        super(XRayTube, self).__init__()

    async def _get_state(self):
        raise AccessorNotImplementedError

    async def _get_voltage(self):
        raise AccessorNotImplementedError

    async def _set_voltage(self, voltage):
        raise AccessorNotImplementedError

    async def _get_current(self):
        raise AccessorNotImplementedError

    async def _set_current(self, current):
        raise AccessorNotImplementedError

    async def _get_power(self):
        return (await self.get_voltage() * await self.get_current()).to(q.W)

    @background
    @check(source='off', target='on')
    async def on(self):
        """
        on()

        Enables the x-ray tube.
        """
        await self._on()

    @background
    @check(source='on', target='off')
    async def off(self):
        """
        off()

        Disables the x-ray tube.
        """
        await self._off()

    async def _on(self):
        """
        Implementation of on().
        """
        raise NotImplementedError

    async def _off(self):
        """
        Implementation of off().
        """
        raise NotImplementedError



"""A gripper can grip and release objects."""

from concert.base import check, State
from concert.coroutines.base import background
from concert.devices.base import Device


class Gripper(Device):

    """Base gripper class."""

    state = State(default='released')

    @background
    @check(source='gripped', target='released')
    async def release(self):
        """
        release()

        Release an object.
        """
        await self._release()

    @background
    @check(source='released', target='gripped')
    async def grip(self):
        """
        grip()

        Grip an object.
        """
        await self._grip()

    async def _release(self):
        """The actual release implementation."""
        raise NotImplementedError

    async def _grip(self):
        """The actual grip implementation."""
        raise NotImplementedError



class Proxy:
    async def setup(self):
        pass

    async def teardown(self):
        pass

    async def wait(self, num):
        pass


class ImageWriter(Proxy):
    async def set_path(self, path):
        raise NotImplementedError



import asyncio
import logging
from concert.config import AIODEBUG
from concert.coroutines.base import background


LOG = logging.getLogger(__name__)


class Result(object):
    """
    The object is callable and when called it becomes a coroutine which accepts
    items and stores them in a variable which allows the user to obtain the
    last stored item at any time point.
    """
    def __init__(self):
        self.result = None

    @background
    async def __call__(self, producer):
        """
        __call__(self, producer)

        When the object is called we store every item in an attribute which
        can be recovered later.
        """
        async for item in producer:
            self.result = item


@background
async def null(producer):
    """
    null(producer)

    A black-hole.
    """
    try:
        async for item in producer:
            pass
    except asyncio.CancelledError:
        LOG.log(AIODEBUG, 'null cancelled')
        raise


class Accumulate(object):
    """Accumulate items in a list or a numpy array if *shape* is given, *dtype* is the data type.
    """

    def __init__(self, shape=None, dtype=None):
        import numpy as np

        self.items = [] if shape is None else np.empty(shape, dtype=dtype)

    @background
    def __call__(self, producer):
        """
        __call__(self, producer)

        Coroutine interface for processing in a pipeline.
        """
        if isinstance(self.items, list):
            return self._process(producer)
        else:
            return self._process_numpy(producer)

    async def _process(self, producer):
        """Stack data into a list."""
        # Clear results from possible previous execution but keep the list in the same place
        del self.items[:]

        async for item in producer:
            self.items.append(item)

    async def _process_numpy(self, producer):
        """Stack data into a numpy array."""
        i = 0

        async for item in producer:
            self.items[i] = item
            i += 1



import os
import numpy as np
from concert.coroutines.base import async_generate
from concert.coroutines.sinks import Accumulate
from concert.experiments.proxies.base import ImageWriter as BaseImageWriter, Proxy
from concert.storage import DummyWalker, DirectoryWalker


class Consumer(Proxy):
    def __init__(self, consumer):
        self._consumer = consumer

    async def consume(self, producer):
        await self._consumer(producer)


class ImageWriter(BaseImageWriter):
    def __init__(self, walker):
        self._walker = walker
        self._path = None

    async def set_path(self, path):
        self._path = path

    async def write_sequence(self, producer):
        """Wrap the walker and write data."""
        walker = DirectoryWalker(
            writer=self._walker.writer,
            dsetname=self._walker.dsetname,
            bytes_per_file=self._walker._bytes_per_file,
            root=self._path
        )

        await walker.create_writer(producer)


class Accumulator(Proxy):
    def __init__(self):
        self._accumulators = {}

    def accumulate(self, producer, acquisition_name, shape=None, dtype=None):
        if acquisition_name not in self._accumulators:
            self._accumulators[acquisition_name] = Accumulate(shape=shape, dtype=dtype)

        return self._accumulators[acquisition_name](producer)

    async def get_items(self, acquisition_name):
        if acquisition_name not in self._accumulators:
            return []
        return self._accumulators[acquisition_name].items

    async def teardown(self):
        self._accumulators = {}


class OnlineReconstruction(Proxy):
    def __init__(self, reco_args, average_normalization=True, walker=None,
                 slice_directory='online-slices'):
        from concert.ext.ufo import GeneralBackprojectManager

        self.manager = GeneralBackprojectManager(
            reco_args,
            average_normalization=average_normalization
        )
        self.walker = walker
        self.slice_directory = slice_directory

    def update_darks(self, producer):
        return self.manager.update_darks(producer)

    def update_flats(self, producer):
        return self.manager.update_flats(producer)

    async def reconstruct(self, producer):
        await self.manager.backproject(producer)
        if self.walker:
            async with self.walker:
                producer = async_generate(self.manager.volume)
                writer = self.walker.create_writer(
                    producer,
                    name=self.slice_directory,
                    dsetname='slice_{:>04}.tif'
                )
            await writer

    async def get_volume(self):
        return self.manager.volume


class PhaseGratingSteppingFourierProcessing(Proxy):
    def __init__(self, experiment, output_directory="contrasts"):
        self._output_directory = output_directory
        self._experiment = experiment
        self._dark_image = None
        self._reference_stepping = []
        self._object_stepping = []
        self.object_intensity = None
        self.object_phase = None
        self.object_visibility = None
        self.reference_intensity = None
        self.reference_phase = None
        self.reference_visibility = None
        self.intensity = None
        self.diff_phase = None
        self.visibility_contrast = None
        self.diff_phase_in_rad = None

    async def process_darks(self, producer):
        """
        Processes dark images. All dark images are averaged.

        :param producer: Dark image producer
        :return:
        """
        self._dark_image = None
        async for item in producer:
            if self._dark_image is None:
                self._dark_image = item.astype(np.float32)
            else:
                self._dark_image += item
        self._dark_image /= await self._experiment.get_num_darks()

    async def process_stepping(self, producer):
        if await self._experiment.get_acquisition("reference_stepping").get_state() == "running":
            current_stepping = "reference"
            self._reference_stepping = []
        elif await self._experiment.get_acquisition("object_stepping").get_state() == "running":
            current_stepping = "object"
            self._object_stepping = []
        else:
            return

        async for item in producer:
            if current_stepping == "reference":
                self._reference_stepping.append(item)
            elif current_stepping == "object":
                self._object_stepping.append(item)
        if await self._experiment.acquisitions[-1].get_state() == "running":
            await self._compare_and_write()

    async def _process_data_and_write(self, stepping, dark_image, name):
        """
        :param stepping: Stepping data (list of images)
        :param dark_image: Averaged dark image
        :param name: Name of the stepping. If set to 'object' or 'reference', the resulting data is
            stored in the corresponding class variables.
        """
        intensity, phase, visibility = await self._process_data(stepping,
                                                                dark_image)
        if name == "object":
            self.object_intensity = intensity
            self.object_phase = phase
            self.object_visibility = visibility
        elif name == "reference":
            self.reference_intensity = intensity
            self.reference_phase = phase
            self.reference_visibility = visibility

        await self._write_single_image(f"{name}_intensity.tif", intensity)
        await self._write_single_image(f"{name}_phase.tif", phase)
        await self._write_single_image(f"{name}_visibility.tif", visibility)

    async def _compare_and_write(self):
        """
        Processes all acquired data and stores the resulting images.
        """
        if self._object_stepping:
            await self._process_data_and_write(self._object_stepping,
                                               self._dark_image,
                                               "object")

        if self._reference_stepping:
            await self._process_data_and_write(self._reference_stepping,
                                               self._dark_image,
                                               "reference")

        if self._reference_stepping and self._object_stepping:
            self.intensity = self.object_intensity / self.reference_intensity
            self.diff_phase = self.object_phase - self.reference_phase
            self.visibility_contrast = self.object_visibility / self.reference_visibility
            await self._write_single_image("intensity_contrast.tif", self.intensity)
            await self._write_single_image("visibility_contrast.tif", self.visibility_contrast)
            await self._write_single_image("differential_phase.tif", self.diff_phase)
            if self._experiment.get_propagation_distance() is not None:
                self.diff_phase_in_rad = (self.diff_phase *
                                          (await self._experiment.get_grating_period() /
                                           await self._experiment.get_propagation_distance()))
                await self._write_single_image("differential_phase_in_rad.tif",
                                               self.diff_phase_in_rad)

    async def _process_data(self, stepping, dark):
        stepping_curve = np.zeros((stepping[0].shape[0],
                                   stepping[0].shape[1],
                                   len(stepping)),
                                  dtype=np.float32)
        for i, step in enumerate(stepping):
            if dark is not None:
                stepping_curve[:, :, i] = step - dark
            else:
                stepping_curve[:, :, i] = step
        fft_object = np.fft.fft(stepping_curve)

        fft_object = fft_object[:, :, (0, await self._experiment.get_num_periods())] / (
                await self._experiment.get_num_periods() *
                await self._experiment.get_num_steps_per_period())
        phase = np.angle(fft_object[:, :, 1])
        visibility = (2. * np.absolute(fft_object[:, :, 1])) / np.real(fft_object[:, :, 0])
        intensity = np.abs(np.real(fft_object[:, :, 0]))
        return intensity, phase, visibility

    async def _write_single_image(self, name, image):
        async with self._experiment.walker:
            file_name = os.path.join(self._experiment.walker.current, name)

        im_writer = self._experiment.walker.writer(file_name, bytes_per_file=0)
        im_writer.write(image)


class AddonError(Exception):
    """Addon errors."""


class OnlineReconstructionError(Exception):
    pass



"""Pumps."""

from concert.base import Quantity, State, check, AccessorNotImplementedError
from concert.coroutines.base import background
from concert.quantities import q
from concert.devices.base import Device


class Pump(Device):

    """A pumping device."""

    state = State(default='standby')
    flow_rate = Quantity(q.l / q.s, help="Flow rate")

    def __init__(self):
        super(Pump, self).__init__()

    @background
    @check(source='standby', target='pumping')
    async def start(self):
        """
        start()

        Start pumping.
        """
        await self._start()

    @background
    @check(source='pumping', target='standby')
    async def stop(self):
        """
        stop()

        Stop pumping.
        """
        await self._stop()

    async def _get_flow_rate(self):
        raise AccessorNotImplementedError

    async def _set_flow_rate(self, flow_rate):
        raise AccessorNotImplementedError

    async def _start(self):
        raise NotImplementedError

    async def _stop(self):
        raise NotImplementedError



from concert.base import Parameter, ParameterError
from concert.devices.base import Device
from concert.tests import TestCase
from concert.devices.scales.dummy import Scales, TarableScales
from concert.devices.pumps.dummy import Pump
from concert.devices.io.dummy import IO
from concert.devices.monochromators.dummy import Monochromator
from concert.devices.shutters.dummy import Shutter
from concert.devices.storagerings.dummy import StorageRing
from concert.devices.photodiodes.dummy import PhotoDiode
from concert.devices.lightsources.dummy import LightSource


class MockDevice(Device):

    readonly = Parameter()
    writeonly = Parameter()

    def __init__(self):
        super(MockDevice, self).__init__()
        self.aborted = False

    async def _get_readonly(self):
        return 1

    async def _set_writeonly(self, value):
        pass

    async def _emergency_stop(self):
        self.aborted = True


class TestDevice(TestCase):

    def setUp(self):
        super(TestDevice, self).setUp()
        self.device = MockDevice()

    def test_dummies(self):
        Scales()
        TarableScales()
        Pump()
        IO()
        Monochromator()
        Shutter()
        StorageRing()
        PhotoDiode()
        LightSource()

    async def test_accessor_functions(self):
        self.assertEqual(await self.device.get_readonly(), 1)
        await self.device.set_writeonly(0)

    def test_iterable(self):
        for param in self.device:
            self.assertTrue(param.name in ('readonly', 'writeonly'))

    def test_invalid_parameter(self):
        def query_invalid_param():
            self.device['foo']

        self.assertRaises(ParameterError, query_invalid_param)

    def test_str(self):
        from concert.session.utils import get_default_table
        table = get_default_table(["Parameter", "Value"])
        table.border = False
        table.add_row(["readonly", "1"])
        table.add_row(["writeonly", "N/A"])
        self.assertEqual(str(self.device), table.get_string())

    async def test_context_manager(self):
        # This is just a functional test, we don't cover synchronization issues
        # here.
        async with self.device as d:
            await d.get_readonly()
            await d.set_writeonly(2)

    async def test_parameter_lock_acquisition(self):
        async with self.device['writeonly']:
            pass

    async def test_emergency_stop(self):
        await self.device.emergency_stop()
        self.assertTrue(self.device.aborted)



"""A dummy detector."""
from concert.devices.detectors import base
from concert.devices.cameras.dummy import Camera


class Detector(base.Detector):

    """A dummy detector."""

    def __init__(self, camera=None, magnification=None):
        self._camera = Camera() if camera is None else camera
        self._magnification = 3 if magnification is None else magnification
        super(Detector, self).__init__()

    @property
    def camera(self):
        return self._camera

    @property
    def magnification(self):
        return self._magnification



"""
An experiment can be run multiple times. The base :py:class:`.Experiment` takes
care of proper logging structure.
"""

import asyncio
import logging
import os
import time
from concert.coroutines.base import background, broadcast
from concert.coroutines.sinks import null
from concert.progressbar import wrap_iterable
from concert.base import check, Parameterizable, Parameter, Selection, State, StateError, transition


LOG = logging.getLogger(__name__)


class Acquisition(object):

    """
    An acquisition acquires data, gets it and sends it to consumers.

    .. py:attribute:: producer

        a callable with no arguments which returns a generator yielding data items once called.

    .. py:attribute:: consumers

        a list of callables with no arguments which return a coroutine consuming the data once
        started, can be empty.

    .. py:attribute:: acquire

        a coroutine function which acquires the data, takes no arguments, can be None.

    """

    def __init__(self, name, producer, consumers=None, acquire=None):
        self.name = name
        self.producer = producer
        self.consumers = [] if consumers is None else consumers
        # Don't bother with checking this for None later
        if acquire and not asyncio.iscoroutinefunction(acquire):
            raise TypeError('acquire must be a coroutine function')
        self.acquire = acquire

    @background
    async def __call__(self):
        """Run the acquisition, i.e. acquire the data and connect the producer and consumers."""
        LOG.debug(f"Running acquisition '{self.name}'")
        consumers = self.consumers
        if not consumers:
            LOG.debug(f"`{self.name}' has no consumers, using null")
            consumers = [null]

        if self.acquire:
            await self.acquire()

        coros = broadcast(self.producer(), *consumers)
        await asyncio.gather(*coros, return_exceptions=False)

    def __repr__(self):
        return "Acquisition({})".format(self.name)


class Experiment(Parameterizable):

    """
    Experiment base class. An experiment can be run multiple times with the output data and log
    stored on disk. You can prepare every run by :meth:`.prepare` and finsh the run by
    :meth:`.finish`. These methods do nothing by default. They can be useful e.g. if you need to
    reinitialize some experiment parts or want to attach some logging output.

    .. py:attribute:: acquisitions
        :noindex:

        A list of acquisitions this experiment is composed of

    .. py:attribute:: walker

       A :class:`concert.storage.Walker` descends to a data set specific for every run if given

    .. py:attribute:: separate_scans

        If True, *walker* does not descend to data sets based on specific runs

    .. py:attribute:: name_fmt

        Since experiment can be run multiple times each iteration will have a separate entry
        on the disk. The entry consists of a name and a number of the current iteration, so the
        parameter is a formattable string.

    """

    iteration = Parameter()
    separate_scans = Parameter()
    name_fmt = Parameter()
    state = State(default='standby')
    log_level = Selection(['critical', 'error', 'warning', 'info', 'debug'])

    def __init__(self, acquisitions, walker=None, separate_scans=True, name_fmt='scan_{:>04}'):
        self._acquisitions = []
        for acquisition in acquisitions:
            self.add(acquisition)
        self.walker = walker
        self._separate_scans = separate_scans
        self._name_fmt = name_fmt
        self._iteration = 1
        self.log = LOG
        super(Experiment, self).__init__()

        if separate_scans and walker:
            # The data is not supposed to be overwritten, so find an iteration which
            # hasn't been used yet
            while self.walker.exists(self._name_fmt.format(self._iteration)):
                self._iteration += 1

    async def _get_iteration(self):
        return self._iteration

    async def _set_iteration(self, iteration):
        self._iteration = iteration

    async def _get_separate_scans(self):
        return self._separate_scans

    async def _set_separate_scans(self, separate_scans):
        self._separate_scans = separate_scans

    async def _get_name_fmt(self):
        return self._name_fmt

    async def _set_name_fmt(self, fmt):
        self._name_fmt = fmt

    async def _get_log_level(self):
        return logging.getLevelName(self.log.getEffectiveLevel()).lower()

    async def _set_log_level(self, level):
        self.log.setLevel(level.upper())

    async def prepare(self):
        """Gets executed before every experiment run."""
        pass

    async def finish(self):
        """Gets executed after every experiment run."""
        pass

    @property
    def acquisitions(self):
        """Acquisitions is a read-only attribute which has to be manipulated by explicit methods
        provided by this class.
        """
        return tuple(self._acquisitions)

    def add(self, acquisition):
        """
        Add *acquisition* to the acquisition list and make it accessible as
        an attribute::

            frames = Acquisition(...)
            experiment.add(frames)
            # This is possible
            experiment.frames
        """
        self._acquisitions.append(acquisition)
        setattr(self, acquisition.name, acquisition)

    def remove(self, acquisition):
        """Remove *acquisition* from experiment."""
        self._acquisitions.remove(acquisition)
        delattr(self, acquisition.name)

    def swap(self, first, second):
        """
        Swap acquisition *first* with *second*. If there are more occurences
        of either of them then the ones which are found first in the acquisitions
        list are swapped.
        """
        if first not in self._acquisitions or second not in self._acquisitions:
            raise ValueError("Both acquisitions must be part of the experiment")

        first_index = self._acquisitions.index(first)
        second_index = self._acquisitions.index(second)
        self._acquisitions[first_index] = second
        self._acquisitions[second_index] = first

    def get_acquisition(self, name):
        """
        Get acquisition by its *name*. In case there are more like it, the first
        one is returned.
        """
        for acq in self._acquisitions:
            if acq.name == name:
                return acq
        raise ExperimentError("Acquisition with name `{}' not found".format(name))

    async def acquire(self):
        """
        Acquire data by running the acquisitions. This is the method which implements
        the data acquisition and should be overriden if more functionality is required,
        unlike :meth:`~.Experiment.run`.
        """
        for acq in wrap_iterable(self._acquisitions):
            if await self.get_state() != 'running':
                break
            await acq()

    @background
    @check(source=['standby', 'error'], target='standby')
    @transition(immediate='running', target='standby')
    async def run(self):
        start_time = time.time()
        handler = None
        iteration = await self.get_iteration()
        separate_scans = await self.get_separate_scans()

        try:
            if self.walker:
                if separate_scans:
                    self.walker.descend((await self.get_name_fmt()).format(iteration))
                if os.path.exists(self.walker.current):
                    # We might have a dummy walker which doesn't create the directory
                    handler = logging.FileHandler(os.path.join(self.walker.current,
                                                               'experiment.log'))
                    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s '
                                                  '- %(message)s')
                    handler.setFormatter(formatter)
                    self.log.addHandler(handler)
            self.log.info(await self.info_table)
            LOG.debug('Experiment iteration %d start', iteration)
            await self.prepare()
            await self.acquire()
        except asyncio.CancelledError:
            # This is normal, no special state needed -> standby
            LOG.warn('Experiment cancelled')
        except Exception as e:
            # Something bad happened and we can't know what, so set the state to error
            LOG.warn(f"Error `{e}' while running experiment")
            raise StateError('error', msg=str(e))
        except KeyboardInterrupt:
            LOG.warn('Experiment cancelled by keyboard interrupt')
            self._state_value = 'standby'
            raise
        finally:
            try:
                await self.finish()
            except Exception as e:
                LOG.warn(f"Error `{e}' while finalizing experiment")
                raise StateError('error', msg=str(e))
            finally:
                if separate_scans and self.walker:
                    self.walker.ascend()
                LOG.debug('Experiment iteration %d duration: %.2f s',
                          iteration, time.time() - start_time)
                if handler:
                    handler.close()
                    self.log.removeHandler(handler)
                await self.set_iteration(iteration + 1)


class AcquisitionError(Exception):
    """Acquisition-related exceptions."""
    pass


class ExperimentError(Exception):
    """Experiment-related exceptions."""
    pass



import asyncio
import logging
from concert.base import Parameterizable
from concert.coroutines.base import background


LOG = logging.getLogger(__name__)


class Device(Parameterizable):

    """
    A :class:`.Device` provides locked access to a real-world device.

    It implements the context protocol to provide locking::

        async with device:
            # device is locked
            await device.set_parameter(1 * q.m)
            ...

        # device is unlocked again
    """

    def __init__(self):
        # We have to create the lock early on because it will be accessed in
        # any add_parameter calls, especially those in the Parameterizable base
        # class
        self._lock = asyncio.Lock()
        super(Device, self).__init__()

    async def __aenter__(self):
        await self._lock.acquire()

        return self

    async def __aexit__(self, exc_type, exc, tb):
        self._lock.release()

    @background
    async def emergency_stop(self):
        """Emergency stop."""
        await self._emergency_stop()

    async def _emergency_stop(self):
        """Emergency stop implementation."""
        pass



import numpy as np
from concert.tests import TestCase, assert_almost_equal
from concert.quantities import q
from concert.devices.motors.dummy import LinearMotor
from concert.devices.positioners.base import PositionerError, Axis
from concert.devices.positioners.dummy import Positioner, ImagingPositioner


ORIGIN = (0.0, 0.0, 0.0) * q.mm
ROT_ORIGIN = (0.0, 0.0, 0.0) * q.rad


class TestAxis(TestCase):

    def setUp(self):
        super(TestAxis, self).setUp()
        self.motor = LinearMotor()
        self.motor.position = 0 * q.mm

    async def test_positive_direction(self):
        # Test positive direction
        axis = Axis('x', self.motor, direction=1)
        await axis.set_position(1 * q.mm)
        assert_almost_equal(await self.motor.get_position(), await axis.get_position())

    async def test_negative_direction(self):
        # Test positive direction
        axis = Axis('x', self.motor, direction=-1)
        await axis.set_position(-1 * q.mm)
        assert_almost_equal(await self.motor.get_position(), - (await axis.get_position()))


class TestPositioners(TestCase):

    def setUp(self):
        super(TestPositioners, self).setUp()
        self.positioner = Positioner()
        self.positioner.position = ORIGIN
        self.positioner.orientation = ROT_ORIGIN

    def test_position(self):
        position = (1.0, 2.0, 3.0) * q.um
        self.positioner.position = position
        assert_almost_equal(position, self.positioner.position)

        # Test non-existent axis
        del self.positioner.translators['x']
        with self.assertRaises(PositionerError):
            self.positioner.position = position

        # The remaining axes must work
        position = (0.0, 1.0, 2.0) * q.mm
        self.positioner.position = position
        assert_almost_equal(position[1:], self.positioner.position[1:])

        # Also nan must work
        position = (np.nan, 1.0, 2.0) * q.mm
        self.positioner.position = position
        assert_almost_equal(position[1:], self.positioner.position[1:])

        # Also 0 in the place of no axis must work
        self.positioner.position = (0.0, 1.0, 2.0) * q.mm
        assert_almost_equal(position[1:], self.positioner.position[1:])

    def test_orientation(self):
        orientation = (1.0, 2.0, 3.0) * q.rad
        self.positioner.orientation = orientation
        assert_almost_equal(orientation, self.positioner.orientation)

        # Degrees must be accepted
        self.positioner.orientation = (2.0, 3.0, 4.0) * q.deg

        # Test non-existent axis
        del self.positioner.rotators['x']
        with self.assertRaises(PositionerError):
            self.positioner.orientation = orientation

        # Also nan must work
        orientation = (np.nan, 1.0, 2.0) * q.rad
        self.positioner.orientation = orientation
        # assert_almost_equal(orientation[1:], self.positioner.orientation[1:])

        # Also 0 in the place of no axis must work
        self.positioner.orientation = (0.0, 1.0, 2.0) * q.rad
        # assert_almost_equal(orientation[1:], self.positioner.orientation[1:])

    async def test_move(self):
        position = (1.0, 2.0, 3.0) * q.mm
        await self.positioner.move(position)
        assert_almost_equal(position, await self.positioner.get_position())

    async def test_rotate(self):
        orientation = (1.0, 2.0, 3.0) * q.rad
        await self.positioner.rotate(orientation)
        assert_almost_equal(orientation, await self.positioner.get_orientation())

    async def test_right(self):
        await self.positioner.right(1 * q.mm)
        assert_almost_equal((1.0, 0.0, 0.0) * q.mm, await self.positioner.get_position())

    async def test_left(self):
        await self.positioner.left(1 * q.mm)
        assert_almost_equal((-1.0, 0.0, 0.0) * q.mm, await self.positioner.get_position())

    async def test_up(self):
        await self.positioner.up(1 * q.mm)
        assert_almost_equal((0.0, 1.0, 0.0) * q.mm, await self.positioner.get_position())

    async def test_down(self):
        await self.positioner.down(1 * q.mm)
        assert_almost_equal((0.0, -1.0, 0.0) * q.mm, await self.positioner.get_position())

    async def test_forward(self):
        await self.positioner.forward(1 * q.mm)
        assert_almost_equal((0.0, 0.0, 1.0) * q.mm, await self.positioner.get_position())

    async def test_back(self):
        await self.positioner.back(1 * q.mm)
        assert_almost_equal((0.0, 0.0, -1.0) * q.mm, await self.positioner.get_position())


class TestImagingPositioner(TestCase):

    def setUp(self):
        super(TestImagingPositioner, self).setUp()
        self.positioner = ImagingPositioner()
        self.positioner.position = ORIGIN
        self.positioner.orientation = ROT_ORIGIN
        self._pixel_width_position = self.positioner.detector.pixel_width.to(q.m).magnitude
        self._pixel_height_position = self.positioner.detector.pixel_height.to(q.m).magnitude

    async def test_move(self):
        pixel_width = (await self.positioner.detector.get_pixel_width()).to(q.m).magnitude
        pixel_height = (await self.positioner.detector.get_pixel_height()).to(q.m).magnitude

        await self.positioner.move((100.0, 200.0, 0.0) * q.pixel)
        position = (100.0 * pixel_width, 200.0 * pixel_height, 0.0) * q.m
        assert_almost_equal(position, await self.positioner.get_position())

        # Cannot move in z-direction by pixel size
        with self.assertRaises(PositionerError):
            await self.positioner.move((1.0, 2.0, 3.0) * q.pixel)

    async def test_right(self):
        await self.positioner.right(1 * q.px)
        assert_almost_equal((self._pixel_width_position, 0.0, 0.0) * q.m,
                            await self.positioner.get_position())

    async def test_left(self):
        await self.positioner.left(1 * q.px)
        assert_almost_equal((- self._pixel_width_position, 0.0, 0.0) * q.m,
                            await self.positioner.get_position())

    async def test_up(self):
        await self.positioner.up(1 * q.px)
        assert_almost_equal((0.0, self._pixel_height_position, 0.0) * q.m,
                            await self.positioner.get_position())

    async def test_down(self):
        await self.positioner.down(1 * q.px)
        assert_almost_equal((0.0, - self._pixel_height_position, 0.0) * q.m,
                            await self.positioner.get_position())

    async def test_forward(self):
        with self.assertRaises(PositionerError):
            await self.positioner.forward(1 * q.px)

    async def test_back(self):
        with self.assertRaises(PositionerError):
            await self.positioner.back(1 * q.px)



"""Connection protocols for network communication."""
import asyncio
import os
import logging
from concert.quantities import q
from concert.config import AIODEBUG


LOG = logging.getLogger(__name__)


class SocketConnection(object):

    """A two-way socket connection. *return_sequence* is a string appended
    after every command indicating the end of it, the default value
    is a newline (\\n).
    """

    def __init__(self, host, port, return_sequence="\n"):
        self._peer = (host, port)
        self._reader = None
        self._writer = None
        self.lock = asyncio.Lock()
        self.return_sequence = return_sequence

    async def __aenter__(self):
        LOG.log(AIODEBUG, 'Socket connection enter')
        await self.connect()

        return self

    async def __aexit__(self, exc_type, exc, tb):
        LOG.log(AIODEBUG, 'Socket connection exit')
        await self.close()

    async def connect(self):
        """Open connection."""
        self._reader, self._writer = await asyncio.open_connection(*self._peer)

    async def close(self):
        """Close connection."""
        self._writer.close()
        await self._writer.wait_closed()

    async def send(self, data):
        """
        Send *data* to the peer. The return sequence characters
        are appended to the data before it is sent.
        """
        if not self._reader:
            await self.connect()

        LOG.debug('Sending {0}'.format(data))
        data += self.return_sequence
        self._writer.write(data.encode('ascii'))
        await self._writer.drain()

    async def recv(self, num=1024):
        """
        Read *num* bytes from the socket. The result is first stripped from the trailing return
        sequence characters and then returned.
        """
        if not self._reader:
            await self.connect()

        result = (await self._reader.read(num)).decode('ascii')
        if result.endswith(self.return_sequence):
            # Strip the command-ending character
            result = result.rstrip(self.return_sequence)
        LOG.debug('Received {0}'.format(result))
        return result

    async def execute(self, data, num=1024):
        """Execute command and wait for response (coroutine-safe, not thread-safe). Read *num* bytes
        from the socket.
        """
        async with self.lock:
            await self.send(data)
            result = await self.recv(num=num)

        return result


def get_tango_device(uri, peer=None, timeout=10 * q.s):
    """
    Get a Tango device by specifying its *uri*. If *peer* is given change the tango_host specifying
    which database to connect to. Format is host:port as a string. *timeout* sets the device's
    general timeout. It is converted to milliseconds, converted to integer and then the tango
    device's `set_timout_millis` is called with the converted integer value.
    """
    import PyTango
    # TODO: check if there is a way to adjust the host in PyTango.
    if peer is not None:
        os.environ["TANGO_HOST"] = peer

    device = PyTango.DeviceProxy(uri, green_mode=PyTango.GreenMode.Asyncio)
    device.set_timeout_millis(int(timeout.to(q.ms).magnitude))

    return device



"""__init__.py"""
__version__ = '0.30.0'


def get_canonical_version(version=None):
    """Return a version without any dev-suffixes"""
    if not version:
        version = __version__

    return version[:-3] if version.endswith('dev') else version


def require(version):
    """
    Check if the *version* string matches the installed version. If it is
    higher than the installed version, a RuntimeError is raised.
    """
    def get_triple(v):
        return tuple([int(x) for x in get_canonical_version(v).split('.')])

    current_triple = get_triple(__version__)
    required_triple = get_triple(version)

    if current_triple < required_triple:
        import warnings
        msg = "Concert {} is required but only version {} is installed"
        warnings.warn(msg.format(version, __version__), RuntimeWarning)



"""Qt Widget."""
try:
    from PyQt4 import QtGui
except ImportError:
    print("PyQt4 is not installed")


def make(device):
    """Create a Qt widget off of a *device*."""
    layout = QtGui.QGridLayout()

    for row, param in enumerate(device):
        name_widget = QtGui.QLabel(str(param.name))
        layout.addWidget(name_widget, row, 0)

        value = param.get().result()
        value_widget = None

        if isinstance(value, str):
            value_widget = QtGui.QLabel()
            value_widget.setText(value)
        else:
            value_widget = QtGui.QDoubleSpinBox()
            value_widget.setValue(value)

        layout.addWidget(value_widget, row, 1)

    return layout



"""Light sources"""
from concert.quantities import q
from concert.base import Quantity
from concert.devices.base import Device
from concert.base import AccessorNotImplementedError


class LightSource(Device):

    """A base LightSource class."""

    intensity = Quantity(q.V)

    def __init__(self):
        super(LightSource, self).__init__()

    async def _set_intensity(self, value):
        raise AccessorNotImplementedError

    async def _get_intensity(self):
        raise AccessorNotImplementedError



"""Dummy photodiode implementation."""
from concert.quantities import q
from concert.devices.photodiodes import base


class PhotoDiode(base.PhotoDiode):

    """A dummy photo diode"""

    def __init__(self):
        super(PhotoDiode, self).__init__()

    async def _get_intensity(self):
        return 1 * q.V / q.W



import asyncio
import logging
import math
import os
import inspect
import subprocess
import prettytable
from concert.config import AIODEBUG
from concert.coroutines.base import background
from concert.devices.base import Device
from concert.quantities import q


LOG = logging.getLogger(__name__)


def _get_param_description_table(device, max_width):
    field_names = ["Name", "Unit", "Description"]
    widths = [int(math.floor(f * max_width)) for f in (0.3, 0.2, 0.5)]
    table = get_default_table(field_names, widths)
    table.border = False
    table.header = True

    for param in device:
        units = param.unit if hasattr(param, 'unit') else None
        row = [param.name, str(units), str(param)]
        table.add_row(row)

    return table.get_string()


def _get_param_value_table(device):
    field_names = ["Name", "Value"]
    table = get_default_table(field_names)
    table.border = False
    table.header = False

    for param in device:
        table.add_row([param.name, str(param.get().result())])

    if hasattr(device, 'state'):
        table.add_row(['state', device.state])

    return table.get_string()


def _current_instances(instance_type):
    # Get the second frame to skip caller who called us
    frame = inspect.stack()[2]
    instances = frame[0].f_globals

    return ((name, obj) for (name, obj)
            in list(instances.items())
            if isinstance(obj, instance_type))


def get_default_table(field_names, widths=None):
    """Return a prettytable styled for use in the shell. *field_names* is a
    list of table header strings."""
    table = prettytable.PrettyTable(field_names)
    table.border = True
    table.hrules = prettytable.ALL
    table.vertical_char = ' '
    table.junction_char = '-'

    def left_align(name):
        # Try different ways of setting the alignment to support older versions
        # of prettytable.
        try:
            table.align[name] = 'l'
        except AttributeError:
            table.set_field_align(name, 'l')

    if widths:
        for name, width in zip(field_names, widths):
            left_align(name)
            table.max_width[name] = width
    else:
        for name in field_names:
            left_align(name)

    return table


def dstate():
    """Render device state in a table."""

    field_names = ["Name", "Parameters"]
    table = get_default_table(field_names)

    for name, device in _current_instances(Device):
        values = _get_param_value_table(device)
        table.add_row([name, values])

    print(table.get_string())


def ddoc():
    """Render device documentation."""
    from concert.devices.base import Device

    n_columns = int(subprocess.check_output(['stty', 'size']).split()[1]) - 8
    field_names = ["Name", "Description", "Parameters"]
    widths = [int(math.floor(f * n_columns)) for f in (0.1, 0.2, 0.7)]
    table = get_default_table(field_names, widths)

    for name, device in _current_instances(Device):
        doc = _get_param_description_table(device, widths[2] - 4)
        table.add_row([name, inspect.getdoc(device), doc])

    print(table.get_string())


def pdoc(hide_blacklisted=True):
    """Render process documentation."""
    black_listed = ('show', 'start', 'init', 'rm', 'log', 'edit', 'import')
    field_names = ["Name", "Description"]
    table = get_default_table(field_names)

    frame = inspect.stack()[1]
    instances = frame[0].f_globals

    for name, obj in list(instances.items()):
        if not name.startswith('_') and inspect.isfunction(obj):
            if hide_blacklisted and name not in black_listed:
                table.add_row([name, inspect.getdoc(obj)])

    print(table.get_string())


def code_of(func):
    """Show implementation of *func*."""
    source = inspect.getsource(func)

    try:
        from pygments import highlight
        from pygments.lexers import PythonLexer
        from pygments.formatters import TerminalFormatter

        print(highlight(source, PythonLexer(), TerminalFormatter()))
    except ImportError:
        print(source)


def abort_awaiting(background=False, skip=None):
    """Abort task currently being awaited in the session. Return True if there is a task being
    awaited, otherwise False. This function does not touch tasks running in the background unless
    *background* is True, in which case it cancels all awaitables.
    """
    import concert.coroutines.base as cbase
    import functools
    import threading

    # Figure out if we are in a callback (ctrl-c or ctrl-k) or check_emergency_stop
    try:
        asyncio.get_running_loop()
    except RuntimeError:
        loop_running = False
    else:
        loop_running = True
    LOG.log(AIODEBUG, 'async code: %s', loop_running)

    def get_task_name(task):
        return task.get_coro().__qualname__

    def get_task_directory(task):
        return os.path.dirname(task.get_coro().cr_code.co_filename)

    if cbase._TLOOP:
        def thread_callback(event, task):
            event.set()

        tasks = asyncio.all_tasks(loop=cbase._TLOOP)
        LOG.log(AIODEBUG, 'Running %d tasks in separate thread:%s', len(tasks),
                '\n'.join([get_task_name(task) for task in tasks]))
        events = []
        for task in tasks:
            events.append(threading.Event())
            task.add_done_callback(functools.partial(thread_callback, events[-1]))
            name = get_task_name(task)
            LOG.log(AIODEBUG, f"Cancelling task `{name}' with result {task.cancel()}")

        # _TLOOP is running here, so we cannot schedule a gathering coroutine, so we create an Event
        # for every coroutine and set it in the add_done_callback and wait for it, which blocks the
        # main thread until everything has been cancelled and completed.
        for event in events:
            event.wait()

    loop = asyncio.get_event_loop()
    try:
        LOG.debug('Global abort called, loop: %d, IPython loop: %d', id(loop),
                  id(get_ipython().pt_loop))
    except NameError:
        LOG.debug('NameError in pid: %d', os.getpid())

    tasks = asyncio.all_tasks(loop=loop)
    pending = []
    LOG.log(AIODEBUG, 'Running %d tasks:\n%s', len(tasks),
            '\n'.join([get_task_name(task) for task in tasks]))

    for task in tasks:
        name = get_task_name(task)
        if skip and name in skip:
            LOG.log(AIODEBUG, 'Skipping task %s', name)
            continue
        abortable = False
        if background and hasattr(task, '_is_concert_task'):
            # ctrl-k, cancel everything from us
            abortable = True
        elif 'run_cell_async' in name:
            # ctrl-c
            # TODO: make this cleaner and robust (run_cell_async may change its name and so on)
            abortable = True
        if abortable:
            # We either abort everything which is enabled for aborting or the current coroutine
            # which is being awaited in the session
            cancelled_result = task.cancel()
            pending.append(task)
            LOG.log(AIODEBUG, "Cancelling task `%s' with result %s", name, cancelled_result)
            if not background:
                return True

    if not loop_running:
        # In case we are called directly in the session
        loop.run_until_complete(asyncio.gather(*pending, return_exceptions=True))

    if background or not loop_running:
        # Either ctrl-k or ctrl-c in a non-async function
        # Use ipython magic because _current_instances won't work here (different stack)
        try:
            uns = get_ipython().user_ns
        except NameError:
            # Invoked from tests
            return False
        devices = [(name, value) for name, value in uns.items() if isinstance(value, Device)]
        emergency_stops = []
        emergency_device_names = []
        for name, value in uns.items():
            if not name.startswith('_') and isinstance(value, Device):
                # ipython adds _1, _2, ... variables to user_ns when people type variable names in
                # the shell
                LOG.info("Emergency stop on `%s'", name)
                emergency_stops.append(value.emergency_stop())
                emergency_device_names.append(name)
        if not loop_running:
            results = loop.run_until_complete(asyncio.gather(*emergency_stops,
                                                             return_exceptions=True))
            for i, result in enumerate(results):
                LOG.log(AIODEBUG, "Emergency stop result for `%s': %s",
                        emergency_device_names[i], result)

    return False


@background
async def check_emergency_stop(check, poll_interval=0.1 * q.s, exit_session=False):
    """
    check_emergency_stop(check, poll_interval=0.1*q.s, exit_session=False)
    If a callable *check* returns True abort is called. Then until it clears to False nothing is
    done and then the process begins again. *poll_interval* is the interval at which *check* is
    called. If *exit_session* is True the session exits when the emergency stop occurs.
    """
    while True:
        if check():
            LOG.error('Emergency stop')
            abort_awaiting(background=True, skip='check_emergency_stop')
            if exit_session:
                os.abort()
            while check():
                # Wait until the flag clears
                await asyncio.sleep(poll_interval.to(q.s).magnitude)
        await asyncio.sleep(poll_interval.to(q.s).magnitude)



import random
from concert.tests import assert_almost_equal, TestCase
from concert.quantities import q
from concert.devices.monochromators.dummy import\
    Monochromator as DummyMonochromator
from concert.devices.monochromators import base
from concert.devices.monochromators.base import Monochromator


class WavelengthMonochromator(Monochromator):

    """
    A monochromator which implements wavelength getter and setter. The
    conversion needs to be handled in the base class.
    """

    def __init__(self):
        super(WavelengthMonochromator, self).__init__()
        self._wavelength = random.random() * 1e-10 * q.m

    async def _get_wavelength_real(self):
        return self._wavelength

    async def _set_wavelength_real(self, wavelength):
        self._wavelength = wavelength


class TestDummyMonochromator(TestCase):

    def setUp(self):
        super(TestDummyMonochromator, self).setUp()
        self.mono = DummyMonochromator()
        self.wave_mono = WavelengthMonochromator()
        self.energy = 25 * q.keV
        self.wavelength = 0.1 * q.nm

    def test_energy_mono_energy(self):
        self.mono.energy = self.energy
        assert_almost_equal(self.mono.energy, self.energy)
        assert_almost_equal(self.mono.wavelength,
                            base.energy_to_wavelength(self.mono.energy))

    def test_energy_mono_wavelength(self):
        self.mono.wavelength = self.wavelength
        assert_almost_equal(self.mono.wavelength, self.wavelength)
        assert_almost_equal(base.wavelength_to_energy(self.wavelength),
                            self.mono.energy)

    def test_wavelength_mono_energy(self):
        self.wave_mono.energy = self.energy
        assert_almost_equal(self.wave_mono.energy, self.energy)
        assert_almost_equal(self.wave_mono.wavelength,
                            base.energy_to_wavelength(self.wave_mono.energy))

    def test_wavelength_mono_wavelength(self):
        # Wavelength-based monochromator.
        self.wave_mono.wavelength = self.wavelength
        assert_almost_equal(self.wave_mono.wavelength, self.wavelength)
        assert_almost_equal(base.wavelength_to_energy(self.wavelength),
                            self.wave_mono.energy)



"""A dummy pump."""

from concert.base import transition
from concert.devices.pumps import base
from concert.quantities import q


class Pump(base.Pump):

    """A dummy pump."""

    def __init__(self):
        super(Pump, self).__init__()
        self._flow_rate = 0 * q.l / q.s

    @transition(target='pumping')
    async def _start(self):
        pass

    @transition(target='standby')
    async def _stop(self):
        pass

    async def _set_flow_rate(self, flow_rate):
        self._flow_rate = flow_rate

    async def _get_flow_rate(self):
        return self._flow_rate



"""A Dummy positioner"""
from concert.devices.motors.dummy import LinearMotor, RotationMotor
from concert.devices.detectors.dummy import Detector
from concert.devices.positioners import base
from concert.devices.positioners.imaging import Positioner as ImagingBasePositioner


def get_axes():
    """Get all the dummy axes."""
    x_lin = LinearMotor()
    y_lin = LinearMotor()
    z_lin = LinearMotor()
    x_rot = RotationMotor()
    y_rot = RotationMotor()
    z_rot = RotationMotor()

    x_lin_axis = base.Axis('x', x_lin)
    y_lin_axis = base.Axis('y', y_lin, direction=-1)
    z_lin_axis = base.Axis('z', z_lin)
    x_rot_axis = base.Axis('x', x_rot)
    y_rot_axis = base.Axis('y', y_rot)
    z_rot_axis = base.Axis('z', z_rot, direction=-1)

    return (x_lin_axis, y_lin_axis, z_lin_axis, x_rot_axis, y_rot_axis, z_rot_axis)


class Positioner(base.Positioner):

    """A dummy positioner."""

    def __init__(self, position=None):
        super(Positioner, self).__init__(get_axes(), position=position)


class ImagingPositioner(ImagingBasePositioner):

    """A dummy imaging positioner."""

    def __init__(self, detector=None, position=None):
        if detector is None:
            detector = Detector()
        super(ImagingPositioner, self).__init__(get_axes(), detector, position=position)



"""
Image processing module for manipulating image data, e.g. filtered
backprojection, flat field correction and other operations on images.
"""

import asyncio
import numpy as np
import logging
from scipy.signal import fftconvolve
from concert.coroutines.base import background, run_in_executor
from concert.quantities import q


LOG = logging.getLogger(__name__)


def normalize(image, minimum=0.0, maximum=1.0):
    """Normalize *image* intensities to start at *minimum* and end at *maximum*."""
    mul = (maximum - minimum) / (image.max() - image.min())

    return mul * (image - image.min()) + minimum


def flat_correct(radio, flat, dark=None):
    """
    Flat field correction of a radiograph *radio* with *flat* field.
    If *dark* field is supplied it is taken into account as well.
    """
    if dark is not None:
        flat = flat - dark
        radio = radio - dark
    valid = np.where(flat != 0)
    result = np.zeros(radio.shape, dtype=np.float32)
    result[valid] = radio[valid] / flat[valid]

    return result


def ramp_filter(width):
    """Get a 1D ramp filter for filtering sinogram rows."""
    base = np.arange(-width // 2, width // 2)

    return np.fft.fftshift(np.abs(base)) * 2.0 / width


@background
async def find_needle_tips(producer):
    """Get sample tips in images from *producer*."""
    tips = []
    coros = []

    async for image in producer:
        # start forces the coroutine to start immediately
        coros.append(run_in_executor(find_needle_tip, image))

    tips = [tip for tip in await asyncio.gather(*coros) if tip is not None]
    LOG.debug('Needle tips: %s', np.array(tips).tolist())

    if len(tips) == 0:
        raise ValueError("No sample tip points found.")

    return tips


def find_needle_tip(image):
    """Extract needle tip from *image*."""
    mask = segment_convex_object(image)
    if mask is None:
        return None
    coords = np.array(list(zip(*np.where(mask))))
    min_y = np.min(coords[:, 0])
    indices = np.where(coords[:, 0] == min_y)[0]
    coords = coords[indices]
    if coords[:, 1].max() - coords[:, 1].min() > image.shape[1] // 4:
        # Needle tip cannot be width / 4 broad, we have probably segmented just noise
        return None
    coords = [_find_peak_subpix(pos, image) for pos in coords]

    return np.mean(coords, axis=0) if coords else None


@background
async def find_sphere_centers_by_mass(producer, border_crossing_ok=True):
    """Get sphere centers in images from *producer* by computing their center of mass. The images
    must be absorption images. If *border_crossing_ok* is False skip images where sphere goes
    outside the field of view.
    """
    def _process_one(image):
        mask = segment_convex_object(image)
        mean_bg = image[mask == 0].mean()
        # Subtract mean of the background to correct for a global grey value offset
        tip = center_of_mass(image - mean_bg)
        if not border_crossing_ok and _touches_border(mask):
            LOG.debug('Skipping border-crossing image with center of mass (x, y) = %s', tip[::-1])
            tip = None

        return tip

    coros = []
    async for image in producer:
        coros.append(run_in_executor(_process_one, image))

    return [tip for tip in await asyncio.gather(*coros) if tip is not None]


@background
async def find_sphere_centers(producer, supersampling=1, correlation_threshold=None):
    """Get sphere centers in images from *producer*.  by finding the image with the largest portion
    of a sphere inside (the sphere may partially go out of the FOV) and correlate other images with
    the found one, from which relative shifts are computed and converted to absolute sphere centers.
    This is done by first computing the center of mass of the best image and then subtracting the
    respective shifts. Use *supersampling* for sub-pixel precision and filter out the centers for
    which the correlation coefficient computed by :func:`.compute_pearson_correlation_coefficient`
    is worse than *correlation_threshold*. The correlation coefficient is computed by shifting an
    image based on the shift found by correlation and computing the correlation coefficient of such
    shifted image with respect to the best one.
    """

    def _wrap(tips, axis):
        t = tips[:, axis]
        indices = np.where(t >= images[0].shape[axis])
        t[indices] = t[indices] - images[0].shape[axis]
        indices = np.where(t < 0)
        t[indices] = t[indices] + images[0].shape[axis]

    def _process_one(image):
        mask = segment_convex_object(image)
        return (image, mask, not _touches_border(mask))

    coros = []
    masks = []
    images = []
    found_completely_in_fov = False
    async for image in producer:
        coros.append(run_in_executor(_process_one, image))

    results = await asyncio.gather(*coros)

    for i, (image, mask, in_fov) in enumerate(results):
        images.append(image)
        masks.append(mask)
        if in_fov:
            a = image
            found_completely_in_fov = True
            LOG.debug('Sphere completely in FOV in image %d', i)

    if not found_completely_in_fov:
        nonzero = [np.count_nonzero(msk) for msk in masks]
        i = np.argmax(nonzero)
        LOG.debug("No sphere commpletely in FOV, largest portion in image %d", i)
        a = images[i]

    center_a = np.mean(np.where(segment_convex_object(a)), axis=1)
    shifts = np.array([correlate(a, b, supersampling=supersampling)[:2] for b in images])
    if correlation_threshold:
        r = np.empty(len(shifts))
        for (i, (dy, dx)) in enumerate(shifts):
            r[i] = await run_in_executor(compute_pearson_correlation_coefficient,
                                         a, images[i], int(np.round(dx)), int(np.round(dy)))
        LOG.debug("Correlation coefficients: %s", r)
        shifts = shifts[np.where(r > correlation_threshold)]

    tips = center_a - shifts
    _wrap(tips, 0)
    _wrap(tips, 1)

    return tips


def segment_convex_object(image):
    """
    Extract convex object from *image* (e.g. needle or sphere). It doesn't matter if object is
    brigher or darker than the background (e.g. non flat corrected radiograph on input).
    """
    try:
        from skimage.filters import threshold_otsu
        from skimage.morphology import dilation, disk, convex_hull_image, label
    except ImportError as e:
        print("You need to install scikit-image in order to use this function")
        LOG.error(e)

    def _segment(threshold, greater=True):
        mask = np.zeros_like(image, dtype=np.int8)
        if greater:
            mask[image > threshold] = 1
        else:
            mask[image < threshold] = 1
        labels, num = label(mask, return_num=True)
        if not num:
            return None
        bins = np.arange(num + 1) + 0.5
        hist, bins = np.histogram(labels, bins=bins)
        largest_label = int(bins[np.argmax(hist)] + 0.5)
        mask[labels != largest_label] = 0

        return mask

    try:
        thr_otsu = threshold_otsu(image)
    except Exception as e:
        LOG.error(e)
        return None

    sgn = 1
    mask = _segment(thr_otsu)
    if mask is None:
        # Nothing found
        return None

    # Compute convex hull of the mask and inverted mask. Object is that mask which has smaller
    # amount of pixels added by the convex hull (convex hull of a convex polygon has the same size
    # as the polygon itself, whereas for concave polygons we'd need to add some pixels to the hull).
    # Since the object segmentation can be jagged, convex hull might still add some pixels, so don't
    # test for 0 but for the difference between added pixels for mask and inverted mask.
    imask = 1 - mask
    hull = convex_hull_image(mask)
    ihull = convex_hull_image(imask)
    hull_diff = np.count_nonzero(hull) - float(np.count_nonzero(mask))
    ihull_diff = np.count_nonzero(ihull) - float(np.count_nonzero(imask))
    if hull_diff > ihull_diff:
        mask = imask
        sgn = -1

    # Refine the segmentation of an object in image by setting the threshold to roughly FWTM of
    # the background standard deviation. Find the background by dilating the mask (in case some
    # object pixels are in the mask) by a small disk and then taking the inverse. sgn controls
    # whether the object is dark or bright (sgn = 1 for bright).
    indices = np.where(1 - dilation(mask, selem=disk(20)))
    if not len(indices[0]):
        return None
    mean_bg = image[indices].mean()
    std_bg = image[indices].std()
    thr = mean_bg + sgn * 5 * std_bg
    if sgn == 1 and thr < thr_otsu or sgn == -1 and thr > thr_otsu:
        mask = _segment(thr, greater=sgn == 1)

    return mask


def _touches_border(mask):
    y, x = np.where(mask)

    return (min(y) == 0 or max(y) == mask.shape[0] - 1
            or min(x) == 0 or max(x) == mask.shape[1] - 1)


def _find_peak_subpix(peak, image, supersampling=16):
    """Supersample vertical line at the *peak* (y, x) position by *supersampling* and look for the
    steepest gradient in the region (peak[0] - 1, peak[0] + 1) in the high resolution line.
    """
    from scipy.ndimage import gaussian_filter1d
    dy = 8
    y_start = max(peak[0] - dy, 0)
    line = image[y_start:min(peak[0] + dy, image.shape[0]), peak[1]]
    x = np.arange(len(line))
    x_hd = np.arange(0, len(line) - 1 + 1. / supersampling, 1. / supersampling)
    line_hd = np.interp(x_hd, x, line)
    # FWHM of the low resolution pixel
    sigma = supersampling / (2. * np.sqrt(2 * np.log(2)))
    blurred = gaussian_filter1d(line_hd, sigma)
    middle = len(x) * supersampling // 2
    g = np.abs(np.gradient(blurred))[middle - supersampling:middle + supersampling + 1]
    y = (np.argmax(g) + middle - supersampling) / supersampling + y_start

    return (y, peak[1])


def _get_boundary_coordinates(coordinates, max_val):
    """Return coordinates which reside on image edges."""
    return [coor for coor in coordinates if coor % max_val == 0]


def _is_corner_point(point, shape):
    """Test if the *point* lies in one of the image corners."""
    return (point[1] == 0 or point[1] == shape[1] - 1) and\
        (point[0] == 0 or point[0] == shape[0] - 1)


def _get_intersection_points(image):
    """Get *image* edges and sample intersection points. The *image* is
    a segmented binary image."""
    y_ind, x_ind = np.where(image != 0)
    x_low = x_ind[np.where(y_ind == 0)]
    x_high = x_ind[np.where(y_ind == image.shape[0] - 1)]
    y_low = y_ind[np.where(x_ind == 0)]
    y_high = y_ind[np.where(x_ind == image.shape[1] - 1)]

    points = []
    if len(x_low) != 0:
        points.append((0, x_low[0]))
        if x_low[-1] != x_low[0]:
            points.append((0, x_low[-1]))
    if len(x_high) != 0:
        points.append((image.shape[0] - 1, x_high[0]))
        if x_high[-1] != x_high[0]:
            points.append((image.shape[0] - 1, x_high[-1]))
    if len(y_low) != 0:
        points.append((y_low[0], 0))
        if y_low[-1] != y_low[0]:
            points.append((y_low[-1], 0))
    if len(y_high) != 0:
        points.append((y_high[0], image.shape[1] - 1))
        if y_high[-1] != y_high[0]:
            points.append((y_high[-1], image.shape[1] - 1))

    if len(points) > 2:
        # The sample is big and besides intersection points it fills some
        # corners of the image.
        res = []
        for point in points:
            if not _is_corner_point(point, image.shape):
                res.append(point)
        points = res

    return points


def _get_axis_intersection(p_1, p_2, shape):
    """Get intersections of a vector perpendicular to a vector defined by
    *p_1* and *p_2* and image edges defined by image *shape*."""
    # First check if the center lies on an edge
    if p_1[0] == p_2[0]:
        return [(p_1[0], (p_1[1] + p_2[1]) / 2)]
    elif p_1[1] == p_2[1]:
        return [((p_1[0] + p_2[0]) / 2, p_1[1])]

    p_x = (p_1[1] + p_2[1]) / 2
    p_y = (p_1[0] + p_2[0]) / 2
    v_y = p_1[0] - p_2[0]
    v_x = p_2[1] - p_1[1]
    height, width = shape[0] - 1, shape[1] - 1

    left = p_y - v_x * p_x / v_y, 0
    right = p_y + v_x * (width - p_x) / v_y, width
    bottom = 0, p_x - v_y * p_y / v_x
    top = height, p_x + v_y * (height - p_y) / v_x

    res = set([left, right, bottom, top])
    # Filter intersections which are out of the image bounding box.
    res = [x for x in res if 0 <= x[0] <= height and 0 <= x[1] <= width]

    return res


def center_of_points(points):
    """
    Find a simplified center of mass withouth point-weighing
    from a set of *points*.
    """
    y_ind, x_ind = list(zip(*points))

    c_y = np.sum(y_ind) / len(points)
    c_x = np.sum(x_ind) / len(points)

    return c_y, c_x


def center_of_mass(frame):
    """Calculates the center of mass of the whole frame wheighted by value."""

    frm_shape = np.array(frame.shape)
    total = frame.sum()
    if total == 0:
        return np.array([-1, -1])
    else:
        y = (frame.sum(1) * np.arange(frm_shape[0])).sum() / total
        x = (frame.sum(0) * np.arange(frm_shape[1])).sum() / total
        return np.array([y, x])


def correlate(first, second, first_y=0, second_y=0, overlap_height=None, supersampling=1):
    """Correlate *first* and *second* image, use *supersampling* for sub-pixel precision. Crop first
    image vertically to (*first_y*, *first_y* + *overlap_height*) and second to (*second_y*,
    *second_y* + *overlap_height*).
    """
    from numpy.fft import fft2, ifft2, fftshift

    try:
        from skimage.filters import sobel
        from skimage.transform import resize
    except ImportError as e:
        print("You need to install scikit-image in order to use this function")
        LOG.error(e)

    height, width = first.shape
    hd_shape = (supersampling * height, supersampling * width)
    first = resize(first, hd_shape, order=1, mode='reflect')
    second = resize(second, hd_shape, order=1, mode='reflect')
    if supersampling > 1:
        ssh = supersampling // 2
        first = first[ssh:-ssh, ssh:-ssh]
        second = second[ssh:-ssh, ssh:-ssh]
    first_y = supersampling * first_y
    if overlap_height:
        overlap_height = supersampling * overlap_height
    else:
        overlap_height = second.shape[0]
    if second_y:
        second_y = supersampling * second_y
    else:
        second_y = second.shape[0] - overlap_height

    first_sobel = sobel(first)[first_y:first_y + overlap_height]
    second_sobel = sobel(second)[second_y:second_y + overlap_height]
    c = fftshift(ifft2(fft2(first_sobel) * np.conjugate(fft2(second_sobel))).real)
    dy, dx = np.unravel_index(c.argmax(), c.shape) - np.array(c.shape) / 2
    dy += second_y - first_y

    return (dy / supersampling, dx / supersampling, c)


def compute_pearson_correlation_coefficient(first, second, dx, dy):
    """Compute Pearson correlation coefficient. Image *second* is shifted by *dx* and *dy* pixels
    and the correlation is computed with respect to image *first*. Both images are cropped with
    respect to the *dx*, *dy* shift in order not to correlate regions overflowing over image edges.
    """
    first = first.copy()
    second = np.roll(np.roll(second.copy(), dy, axis=0), dx, axis=1)
    height, width = first.shape

    if abs(dx) >= width or abs(dy) >= height:
        r = 0
    else:
        # Both images must be cropped with respect to found dx and dy in order not to take garbage
        # into account.
        x_low = max(0, dx)
        x_high = dx if dx < 0 else None
        y_low = max(0, dy)
        y_high = dy if dy < 0 else None
        first = first[y_low:y_high, x_low:x_high]
        first -= first.mean()
        second = second.copy()[y_low:y_high, x_low:x_high]
        second -= second.mean()
        first_std = first.std()
        second_std = second.std()
        if not (first_std and second_std):
            r = 0
        else:
            r = np.mean(first * second) / (first_std * second_std)

    return r


def compute_rotation_axis(first_projection, last_projection):
    """
    Compute the tomographic rotation axis based on cross-correlation technique.
    *first_projection* is the projection at 0 deg, *last_projection* is the projection
    at 180 deg.
    """
    width = first_projection.shape[1]
    first_projection = first_projection - first_projection.mean()
    last_projection = last_projection - last_projection.mean()

    # The rotation by 180 deg flips the image horizontally, in order
    # to do cross-correlation by convolution we must also flip it
    # vertically, so the image is transposed and we can apply convolution
    # which will act as cross-correlation
    convolved = fftconvolve(first_projection, last_projection[::-1, :], mode='same')
    center = np.unravel_index(convolved.argmax(), convolved.shape)[1]

    return (width / 2 + center) / 2 * q.px


def filter_low_frequencies(data, fwhm=32.):
    """Filter low frequencies in 1D *data*. *fwhm* is the FWHM of the gaussian used to filter out
    low frequencies in real space. The window is then computed as fft(1 - gauss).
    """
    mean = np.mean(data)
    sigma = fwhm / (2 * np.sqrt(2 * np.log(2)))
    # We compute the gaussian in Fourier space, so convert sigma first
    f_sigma = 1. / (2 * np.pi * sigma)
    x = np.fft.fftfreq(len(data))
    fltr = 1 - np.exp(- x ** 2 / (2 * f_sigma ** 2))

    return np.fft.ifft(np.fft.fft(data) * fltr).real + mean



"""Port, IO Device."""

import asyncio
from concert.base import AccessorNotImplementedError, State, check
from concert.coroutines.base import background
from concert.quantities import q
from concert.devices.base import Device


class Signal(Device):

    """Base device for binary signals, e.g. TTL trigger signals and similar."""

    state = State(default='off')

    @background
    @check(source='off', target='on')
    async def on(self):
        """
        on()

        Switch the signal on.
        """
        await self._on()

    @background
    @check(source='on', target='off')
    async def off(self):
        """
        off()

        Switch the signal off.
        """
        await self._off()

    @background
    @check(source='off', target='off')
    async def trigger(self, duration=10 * q.ms):
        """
        trigger(duration=10*q.ms)

        Generate a trigger signal of *duration*.
        """
        await self.on()
        await asyncio.sleep(duration.to(q.s).magnitude)
        await self.off()

    async def _on(self):
        """Implementation."""
        raise NotImplementedError

    async def _off(self):
        """Implementation."""
        raise NotImplementedError


class IO(Device):

    """The IO device consists of ports which can be readable, writable or
    both.
    """

    def __init__(self):
        super(IO, self).__init__()
        self._ports = []

    @property
    def ports(self):
        """Port IDs used by :meth:`.read_port` and :meth:`.write_port`"""
        return self._ports

    def _check(self, port):
        """Check if *port* is a part of the device."""
        if port not in self._ports:
            raise IODeviceError("Port `{}' not found".format(port))

    @background
    async def read_port(self, port):
        """Read a *port*."""
        self._check(port)
        return await self._read_port(port)

    @background
    async def write_port(self, port, value):
        """Write a *value* to the *port*."""
        self._check(port)
        await self._write_port(port, value)

    async def _read_port(self, port):
        """Implementation of reading a *port* from the device."""
        raise AccessorNotImplementedError

    async def _write_port(self, port, value):
        """Implementation of writing a *value* to a *port* on the device."""
        raise AccessorNotImplementedError


class IODeviceError(Exception):

    """Specific exception thrown by IO devices."""

    pass



"""
Base class for complex motion realized by combination of more motors
from :mod:`concert.motors.motors.base`.

The :class:`.Positioner` class facilitates multiple :class:`.Axis` for
translation and rotation which form the motion possibilities
of a particular positioner. The coordinate system of the axes within
a positioner is the following:

    * x - horizontal coordinate in the image plane
    * y - vertical coordinate in the image plane
    * z - beam direction

The direction of movements is as follows:

    * + - right/up/forward in x/y/z
    * - - left/down/back in x/y/z

*Note*: *forward* in 'z' directions means away from the detector, one can
interpret it as by "increasing the propagation distance".

Axes which represent angular motion are also represented by a coordinate,
when "y" angular axis means we can rotate around "y" axis. We consider
the system to be "left handed", i.e. that for example the "y" axis
positive rotation is in the clockwise direction.

An exmaple of a positioner which is capable of translation in all
coordinates and rotation around "y"::

    import numpy as np
    from concert.devices.motors.dummy import LinearMotor, RotationMotor

    x_axis = Axis('x', LinearMotor())
    y_axis = Axis('y', LinearMotor())
    z_axis = Axis('z', LinearMotor())
    y_rot_axis = Axis('y', RotationMotor())
    positioner = Positioner([x_axis, y_axis, z_axis, y_rot_axis])

    # Move to the center of origin
    positioner.position = (0, 0, 0) * q.mm
    # Move to the right
    positioner.position = (10, 0, 0) * q.mm
    # Rotate 90 degrees counter clockwise
    positioner.orientation = (np.nan, - 90, np.nan) * q.deg

"""
import asyncio
import numpy as np
from concert.coroutines.base import background
from concert.quantities import q
from concert.base import Quantity
from concert.devices.base import Device
from concert.devices.motors.base import LinearMotor


INF_VECTOR = np.array((np.inf, np.inf, np.inf))


class Axis(object):
    """
    An axis represents a Euclidean axis along which one can translate or
    around which one can rotate. The axis *coordinate* is a string representing
    the Euclidean axis, i.e. 'x' or 'y' or 'z'. Movement is realized by a *motor*.
    An additional *position* argument is necessary for calculatin more complicated
    motion types, e.g. rotation around arbitrary point in space. It is the local
    position with respect to a :class:`concert.devices.positioners.base.Positioner`
    in which it is placed.
    """
    def __init__(self, coordinate, motor, direction=1, position=None):
        self.coordinate = coordinate
        self.motor = motor
        self.direction = direction
        self.position = position

    @background
    async def get_position(self):
        """
        get_position()

        Get position asynchronously with respect to axis direction.
        """
        return (await self.motor.get_position()) * self.direction

    @background
    async def set_position(self, position):
        """Set the *position* asynchronously with respect to axis direction."""
        return await self.motor.set_position(position * self.direction)


class Positioner(Device):
    """Combines more motors which move to form a complex motion. *axes* is a list
    of :class:`.Axis` instances. *position* is a 3D vector of coordinates specifying
    the global position of the positioner.

    If a certain coordinate in the positioner is missing, then when we set the
    position or orientation we can specify the respective vector position to
    be zero or numpy.nan.
    """

    position = Quantity(q.m, help="Global position",
                        lower=-INF_VECTOR * q.m, upper=INF_VECTOR * q.m)

    orientation = Quantity(q.rad, help="Orientation of the coordinate system",
                           lower=-INF_VECTOR * q.rad, upper=INF_VECTOR * q.rad)

    def __init__(self, axes, position=None):
        super(Positioner, self).__init__()
        self.translators = {}
        self.rotators = {}
        self.global_position = None

        for axis in axes:
            if isinstance(axis.motor, LinearMotor):
                self.translators[axis.coordinate] = axis
            else:
                self.rotators[axis.coordinate] = axis

    @background
    async def move(self, position):
        """
        move(position)

        Move by specified *position*.
        """
        await self.set_position(await self.get_position() + position)

    @background
    async def rotate(self, angles):
        """
        rotate(angles)

        Rotate by *angles*.
        """
        await self.set_orientation(await self.get_orientation() + angles)

    @background
    async def right(self, value):
        """
        right(value)

        Move right by *value*."""
        return await self.move(_vectorize(value, 'x'))

    @background
    async def left(self, value):
        """
        left(value)

        Move left by *value*."""
        return await self.right(-value)

    @background
    async def up(self, value):
        """
        up(value)

        Move up by *value*.
        """
        return await self.move(_vectorize(value, 'y'))

    @background
    async def down(self, value):
        """
        down(value)

        Move down by *value*.
        """
        return await self.up(-value)

    @background
    async def forward(self, value):
        """
        forward(value)

        Move forward by *value*.
        """
        return await self.move(_vectorize(value, 'z'))

    @background
    async def back(self, value):
        """
        back(value)

        Move back by *value*.
        """
        return await self.forward(-value)

    async def _get_position(self):
        """Get the position of the positioner."""
        return await self._get_vector(self.translators)

    async def _set_position(self, position):
        """3D translation to *position*."""
        await self._set_vector(position, self.translators)

    async def _get_orientation(self):
        """Get the angular position of the positioner."""
        return await self._get_vector(self.rotators)

    async def _set_orientation(self, angles):
        """Rotation with magnitudes *angles*."""
        await self._set_vector(angles, self.rotators)

    async def _get_vector(self, axes):
        """Get the current translation or orientation vector."""
        vector = []
        unit = q.m if axes == self.translators else q.rad

        for coordinate in ['x', 'y', 'z']:
            if coordinate in axes:
                vector.append((await axes[coordinate].get_position()).to(unit).magnitude)
            else:
                vector.append(np.nan)

        return vector * unit

    async def _set_vector(self, vector, axes):
        """Set position (angular or translational) given by *vector* on *axes*."""
        coros = []
        for i, coordinate in enumerate(['x', 'y', 'z']):
            magnitude = vector[i].magnitude
            if not np.isnan(magnitude):
                if coordinate not in axes:
                    if magnitude != 0:
                        # Last chance is to specify the coordinate to be zero.
                        raise PositionerError('Cannot move in {} coordinate'.format(coordinate))
                else:
                    coro = axes[coordinate].set_position(vector[i])
                    coros.append(coro)

        await asyncio.gather(*coros)


class PositionerError(Exception):
    """Positioning exceptions"""
    pass


def _vectorize(scalar, coordinate):
    """
    Return a vector with the *scalar* in the correct place given by the
    *coordinate* as x, y or z.
    """
    vector = np.zeros(3, dtype=np.float)
    translate = dict(x=0, y=1, z=2)
    vector[translate[coordinate]] = scalar.magnitude

    return q.Quantity(vector, scalar.units)



import time
import inspect
import functools
import logging
from dataclasses import dataclass, field
from typing import Any


LOG = logging.getLogger(__name__)


class Bunch(object):

    """Encapsulate a list or dictionary to provide attribute-like access.

    Common use cases look like this::

        d = {'foo': 123, 'bar': 'baz'}
        b = Bunch(d)
        print(b.foo)
        >>> 123

        l = ['foo', 'bar']
        b = Bunch(l)
        print(b.foo)
        >>> 'foo'
    """

    def __init__(self, values):
        if isinstance(values, list):
            values = dict(list(zip(values, values)))
        self.__dict__.update(values)


def memoize(func):
    """
    Memoize the result of *func*.

    Remember the result of *func* depending on its arguments. Note, that this
    requires that the function is free from any side effects, e.g. returns the
    same value given the same arguments.
    """
    memo = {}

    if inspect.iscoroutinefunction(func):
        async def wrapper(*args):
            if args in memo:
                return memo[args]

            result = await func(*args)
            memo[args] = result
            return result
    else:
        def wrapper(*args):
            if args in memo:
                return memo[args]

            result = func(*args)
            memo[args] = result
            return result

    return wrapper


def measure(func=None, return_result=False):
    """
    Measure and print execution time of *func*.

    If *return_result* is True, the decorated function returns a tuple
    consisting of the original return value and the measured time in seconds.
    """

    if func is not None:
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            start = time.time()
            result = func(*args, **kwargs)
            elapsed = time.time() - start

            if return_result:
                from concert.quantities import q
                return (result, elapsed * q.s)
            else:
                print("`{}' took {}s".format(func.__name__, elapsed))
                return result
        return wrapper
    else:
        return functools.partial(measure, return_result=return_result)


class _Structure(object):

    def __init__(self, func, e_args, f_args, f_defaults, e_keywords):
        self.func = func
        self.e_args = e_args
        self.f_args = f_args
        self.f_defaults = f_defaults
        self.e_keywords = e_keywords
        self._isfunction = True

    def __call__(self, *args, **kwargs):
        self._check_args(*args, **kwargs)
        return self.func(*args, **kwargs)

    def _check_args(self, *args, **kwargs):
        for i, key in enumerate(args):
            if i < len(self.e_args):
                self._check_type_correctness(
                    self.f_args[i],
                    self.e_args[i],
                    args[i])
            else:
                self._check_type_correctness(
                    self.f_args[i],
                    self.e_keywords[self.f_args[i]],
                    args[i])
        for i, key in enumerate(kwargs):
            if kwargs[key] is not None:
                self._check_type_correctness(
                    key,
                    self.e_keywords[key],
                    kwargs[key])

    def _check_type_correctness(self, arg_name, expected, given):
        from concert.devices.base import Device
        import inspect
        if expected is not None:
            if isinstance(expected, Numeric):
                self._check_numeric(arg_name, expected, given)

            elif inspect.isclass(expected):
                if issubclass(expected, Device) and not isinstance(given, expected):
                    raise TypeError(
                        'Sorry, argument "{}" expected to get {}, but got {}'.format(
                            arg_name,
                            expected.__name__,
                            given.__class__.__name__))

    def _check_numeric(self, arg_name, expected, given):
        if (expected.units is not None) ^ hasattr(given, 'units'):
            raise TypeError(
                'Sorry, argument "{}" expected to get value with unit {}, but got {}'.format(
                    arg_name,
                    expected.units,
                    given))

        elif expected.units is not None:
            dim_units = expected.dimension * expected.units
            e_units = dim_units.to_base_units().units
            if not e_units == given.to_base_units().units:
                raise TypeError(
                    'Sorry, argument "{}" expected to get value with unit {}, but got {}'.format(
                        arg_name,
                        expected.units,
                        given.units))

        if hasattr(given, 'magnitude'):
            magnitude = given.magnitude
        else:
            magnitude = given
        if isinstance(magnitude, (float, int)):
            shape = 1
        else:
            shape = len(magnitude)
        if not shape == expected.dimension:
            raise TypeError(
                'Argument {} expected to get value with dimension {}, but got dimension {}'.format(
                    arg_name,
                    expected.dimension,
                    shape))


class expects(object):

    """
    Decorator which determines expected arguments for the function
    and also check correctness of given arguments. If input arguments differ from
    expected ones, exception *TypeError* will be raised.

    For numeric arguments use *Numeric* class with 2 parameters: dimension of
    the array and units (optional). E.g. "Numeric (1)" means function expects
    one number or "Numeric (2, q.mm)" means function expects expression like
    [4,5]*q.mm

    Common use case looks like this::

        from concert.helpers import Numeric

        @expects(Camera, LinearMotor, pixelsize = Numeric(2, q.mm))
        def foo(camera, motor, pixelsize = None):
            pass
    """

    def __init__(self, *args, **kwargs):
        self.e_args = args
        self.e_keywords = kwargs

    def __call__(self, f):
        f_args = inspect.getargspec(f).args
        f_defaults = inspect.getargspec(f).defaults
        self.func = _Structure(
            f,
            self.e_args,
            f_args,
            f_defaults,
            self.e_keywords)
        functools.update_wrapper(self.func, f)
        return self.func


class Numeric(object):
    __name__ = "Numeric"

    def __init__(self, dimension, units=None):
        self.dimension = dimension
        self.units = units


@dataclass(order=True)
class PrioItem:

    """To be used in combination with `queue.PriorityQueue`."""

    priority: int
    data: Any = field(compare=False)


def is_iterable(item):
    """Is *item* iterable or not."""
    try:
        iter(item)
        return True
    except Exception:
        return False



"""
NeXus file format related utitilities.

This module provides convenience utilities to create NeXus-compatible data
sets. It uses NeXpy_ to interface with NeXus.

.. _NeXpy: http://wiki.nexusformat.org/NeXpy
"""
from logging import StreamHandler
from concert.coroutines.base import background
from concert.storage import Walker, StorageError

try:
    import h5py
except ImportError:
    print("h5py is not installed")


class Hdf5Walker(Walker):

    """An HDF5 file walker implementation."""

    def __init__(self, hdf5, dsetname='frames', log=None, log_name='log'):
        """
        *hdf5* is a writeable h5py.File file. *fname* is the dataset name that the
        sequence is stored in.
        """
        log_handler = None
        if log:
            dtype = h5py.special_dtype(vlen=bytes)
            # Append to log if existing
            if log_name not in hdf5:
                log_dset = hdf5.create_dataset(log_name, (1,), maxshape=(None,), dtype=dtype)
            else:
                log_dset = hdf5[log_name]
            log_handler = StreamHandler(stream=Hdf5Stream(log_dset))

        super(Hdf5Walker, self).__init__(hdf5, dsetname=dsetname, log=log, log_handler=log_handler)

    def _descend(self, name):
        if self.exists(name):
            self._current = self._current[name]
        else:
            self._current = self._current.create_group(name)

    def _ascend(self):
        self._current = self._current.parent

    def exists(self, *paths):
        """Check if *paths* exist."""
        return '/'.join(paths) in self.current

    @background
    async def write(self, producer, dsetname=None):
        """Write frames from *producer* to data set *dsetname*."""
        i = 0

        async for data in producer:
            shape = (i + 1, ) + data.shape
            if i == 0:
                maxshape = (None, ) + data.shape
                dsetname = dsetname or self.dsetname

                if dsetname in self._current:
                    raise StorageError("`{}' is not empty".format(self._current.name + '/'
                                                                  + dsetname))

                dset = self._current.create_dataset(dsetname, shape, maxshape=maxshape,
                                                    dtype=data.dtype)
            else:
                dset.resize(shape)
            dset[i, :, :] = data[:, :]
            i += 1


class Hdf5Stream(object):
    """A stream for logging output into an hdf5 file. *dataset* is the dataset of the file to which
    the log output will be written as plain text. The dataset must be resizable and accept plain
    text.
    """

    def __init__(self, dataset):
        """Constructor."""
        self._dset = dataset
        self._iteration = dataset.shape[0] - 1

    def write(self, data):
        """Write logging *data*."""
        self._dset[self._iteration] = data
        self._iteration += 1
        self._dset.resize((self._iteration + 1,))

    def flush(self):
        """flush the stream to the file."""
        self._dset.file.flush()



"""A dummy gripper."""

from concert.devices.grippers import base


class Gripper(base.Gripper):

    """A dummy gripper."""

    def __init__(self):
        super(Gripper, self).__init__()
        self._state = 'released'

    async def _release(self):
        self._state = 'released'

    async def _grip(self):
        self._state = 'gripped'

    async def _get_state(self):
        return self._state



import numpy as np
import scipy
from scipy.ndimage import fourier
from concert.tests import assert_almost_equal, TestCase
from concert.quantities import q
from concert.devices.cameras.dummy import Base as DummyCameraBase, Camera
from concert.devices.motors.dummy import LinearMotor, RotationMotor
from concert.devices.shutters.dummy import Shutter
from concert.processes.common import focus
from concert.processes.beamline import (acquire_dark, acquire_image_with_beam,
                                        determine_rotation_axis)


MIN_POSITION = 0 * q.mm
MAX_POSITION = 100 * q.mm
FOCUS_POSITION = 35 * q.mm


class BlurringCamera(DummyCameraBase):

    def __init__(self, motor):
        super(BlurringCamera, self).__init__()
        self._original = scipy.misc.ascent()
        self.motor = motor

    async def _grab_real(self):
        sigma = abs((await self.motor.get_position() - FOCUS_POSITION).magnitude)
        return fourier.fourier_gaussian(self._original, sigma)


class TestProcesses(TestCase):

    def setUp(self):
        self.motor = LinearMotor()
        self.camera = Camera()
        self.shutter = Shutter()

    async def test_focusing(self):
        await self.motor.set_position(40. * q.mm)
        camera = BlurringCamera(self.motor)
        await focus(camera, self.motor)
        assert_almost_equal(await self.motor.get_position(), FOCUS_POSITION, 1e-2)

    async def test_acquire_dark(self):
        self.assertTrue(isinstance(await acquire_dark(self.camera, self.shutter), np.ndarray))

    async def test_acquire_image_with_beam(self):
        frame = await acquire_image_with_beam(self.camera, self.shutter, self.motor, 1 * q.mm)
        self.assertTrue(isinstance(frame, np.ndarray))
        self.assertEqual(await self.motor.get_position(), 1 * q.mm)

    async def test_determine_rotation_axis(self):
        rot_motor = RotationMotor()
        axis = await determine_rotation_axis(self.camera, self.shutter, self.motor, rot_motor,
                                             1 * q.mm, 3 * q.mm)
        self.assertTrue(isinstance(axis, q.Quantity))



'''
Dummy attenuator box
'''
from concert.devices.attenuatorboxes import base
from concert.base import check, Selection


class AttenuatorBox(base.AttenuatorBox):
    '''Attenuator class implementation'''
    attenuator = Selection([None, 'Al_1mm'], check=check(source='standby', target='standby'))

    def __init__(self):
        super(AttenuatorBox, self).__init__()
        self._filter = None

    async def _set_attenuator(self, attenuator):
        self._filter = attenuator

    async def _get_attenuator(self):
        return self._filter



import asyncio
from concert.tests import TestCase
from concert.base import (transition, check, StateError, TransitionNotAllowed, State, Parameter,
                          FSMError)
from concert.quantities import q
from concert.devices.base import Device


STOP_VELOCITY = 0 * q.mm / q.s
MOVE_VELOCITY = 1 * q.mm / q.s


class SomeDevice(Device):

    state = State(default='standby')

    def __init__(self):
        super(SomeDevice, self).__init__()
        self.velocity = STOP_VELOCITY
        self.faulty = False
        self._error = False

    @transition(target='standby')
    async def make_transition(self):
        if self.faulty:
            raise RuntimeError

    @check(source='standby', target='moving')
    @transition(target='moving')
    async def start_moving(self, velocity):
        self.velocity = velocity

    @check(source='*', target='standby')
    @transition(target='standby')
    async def stop_moving(self):
        self.velocity = STOP_VELOCITY

    @check(source='standby', target='standby')
    @transition(immediate='moving', target='standby')
    async def move_some_time(self, velocity, duration):
        self.velocity = velocity
        await asyncio.sleep(duration)
        self.velocity = STOP_VELOCITY

    @check(source=['standby', 'moving'], target='standby')
    @transition(target='standby')
    async def stop_no_matter_what(self):
        self.velocity = STOP_VELOCITY

    async def _get_state(self):
        if self.faulty:
            raise RuntimeError
        if self._error:
            return 'error'
        return 'standby' if not self.velocity else 'moving'

    @check(source='*', target=['standby', 'moving'])
    async def set_velocity(self, velocity):
        self.velocity = velocity

    @check(source='*', target=['ok', 'error'])
    async def make_error(self):
        self._error = True
        raise StateError('error')

    @check(source='error', target='standby')
    @transition(target='standby')
    async def reset(self):
        self._error = False


class BaseDevice(Device):

    state = State(default='standby')

    def __init__(self):
        super(BaseDevice, self).__init__()

    @check(source='standby', target='in-base')
    @transition(target='in-base')
    async def switch_base(self):
        pass

    @check(source='*', target='standby')
    @transition(target='standby')
    async def reset(self):
        pass


class DerivedDevice(BaseDevice):
    state = State(default='standby')

    def __init__(self):
        super(DerivedDevice, self).__init__()

    @check(source='standby', target='in-derived')
    @transition(target='in-derived')
    async def switch_derived(self):
        pass


class StatelessDevice(Device):
    foo = Parameter(check=check(source='*', target='*'))

    @transition(target='changed')
    async def change(self):
        pass


class TestStateMachine(TestCase):

    def setUp(self):
        super(TestStateMachine, self).setUp()
        self.device = SomeDevice()

    async def test_state_per_device(self):
        d1 = SomeDevice()
        d2 = SomeDevice()

        await d2.start_moving(MOVE_VELOCITY)
        self.assertNotEqual(await d1.get_state(), await d2.get_state())

    async def test_defaults(self):
        await self.device.get_state() == 'standby'

    async def test_valid_transition(self):
        await self.device.start_moving(MOVE_VELOCITY)
        self.assertEqual(await self.device.get_state(), 'moving')

        await self.device.stop_moving()
        self.assertEqual(await self.device.get_state(), 'standby')

    async def test_casync_transition(self):
        await self.device.move_some_time(MOVE_VELOCITY, 0.01)
        self.assertEqual(await self.device.get_state(), 'standby')

    async def test_multiple_source_states(self):
        await self.device.stop_no_matter_what()
        self.assertEqual(await self.device.get_state(), 'standby')

        await self.device.start_moving(MOVE_VELOCITY)
        await self.device.stop_no_matter_what()

        self.assertEqual(await self.device.get_state(), 'standby')

    async def test_multiple_target_states(self):
        await self.device.set_velocity(MOVE_VELOCITY)
        self.assertEqual(await self.device.get_state(), 'moving')

        await self.device.set_velocity(STOP_VELOCITY)
        self.assertEqual(await self.device.get_state(), 'standby')

    async def test_inheritance(self):
        device = DerivedDevice()

        await device.switch_base()
        self.assertEqual(await device.get_state(), 'in-base')

        with self.assertRaises(TransitionNotAllowed):
            await device.switch_base()

        with self.assertRaises(TransitionNotAllowed):
            await device.switch_derived()

        await device.reset()
        self.assertEqual(await device.get_state(), 'standby')

        await device.switch_derived()
        self.assertEqual(await device.get_state(), 'in-derived')

        with self.assertRaises(TransitionNotAllowed):
            await device.switch_derived()

        with self.assertRaises(TransitionNotAllowed):
            await device.switch_base()

        await device.reset()
        self.assertEqual(await device.get_state(), 'standby')

    async def test_errors(self):
        with self.assertRaises(StateError):
            await self.device.make_error()

        self.assertEqual(await self.device.get_state(), 'error')
        await self.device.reset()
        self.assertEqual(await self.device.get_state(), 'standby')

    async def test_state_setting(self):
        with self.assertRaises(AttributeError):
            self.device.state = 'foo'

    def test_stateless_parameter_transition(self):
        dev = StatelessDevice()
        with self.assertRaises(FSMError):
            dev.foo = 1

    async def test_stateless_transition(self):
        dev = StatelessDevice()
        with self.assertRaises(FSMError):
            await dev.change()

    async def test_transition_exception(self):
        dev = SomeDevice()
        dev.faulty = True
        with self.assertRaises(RuntimeError):
            await dev.make_transition()

    async def test_check_exception(self):
        dev = SomeDevice()
        dev.faulty = True
        with self.assertRaises(RuntimeError):
            await dev.start_moving()



import inspect
import time
from concert.tests import TestCase, suppressed_logging
from concert.quantities import q
from concert.helpers import is_iterable, measure, memoize
from concert.processes.common import focus, align_rotation_axis, ProcessError
from concert.devices.motors.dummy import LinearMotor, RotationMotor
from concert.devices.cameras.dummy import Camera


class TestExpects(TestCase):

    def setUp(self):
        self.seq = list(range(10))
        self.camera = Camera()
        self.linear_motor = LinearMotor()
        self.linear_motor2 = LinearMotor()
        self.rotation_motor = RotationMotor()

    async def test_focus_func_arguments_type_error(self):
        with self.assertRaises(TypeError):
            await focus(self.camera, self.rotation_motor)

    async def test_focus_function_arguments(self):
        await focus(self.camera, self.linear_motor)

    async def test_align_rotation_axis_func_type_error(self):
        with self.assertRaises(TypeError):
            await align_rotation_axis(self.camera, self.linear_motor).result()
        with self.assertRaises(TypeError):
            await align_rotation_axis(
                self.camera, self.rotation_motor, self.linear_motor).result()
        with self.assertRaises(TypeError):
            await align_rotation_axis(
                self.camera,
                self.rotation_motor,
                self.rotation_motor,
                self.rotation_motor,
                num_frames=[
                    10,
                    20]).result()
        with self.assertRaises(TypeError):
            await align_rotation_axis(
                self.camera,
                self.rotation_motor,
                self.rotation_motor,
                self.rotation_motor,
                num_frames=10 * q.m
            ).result()

    async def test_align_rotation_axis_function(self):
        with self.assertRaises(ProcessError):
            # Dummy camera, so no tips in noise
            await align_rotation_axis(self.camera, self.rotation_motor, x_motor=self.rotation_motor)


@suppressed_logging
def test_measure_execution():
    @measure(return_result=True)
    def sleeping():
        time.sleep(0.001)
        return 123

    result, elapsed = sleeping()
    assert(result == 123)
    assert(elapsed > 0.001 * q.s)


@suppressed_logging
def test_is_iterable():
    iterables = [(1, 2), [1, 2], {1, 2}]
    noniterables = [1, None, 1 * q.mm]

    # Standard stuff
    for item in iterables:
        assert is_iterable(item)

    # array * unit
    for item in iterables[:2]:
        item = item * q.mm
        assert is_iterable(item)

    # item * unit, item * unit
    assert is_iterable([1 * q.mm, 2 * q.mm])

    for item in noniterables:
        assert not is_iterable(item)


class TestMemoize(TestCase):
    def test_ordinary_func(self):
        ran = False

        @memoize
        def func(arg):
            nonlocal ran
            ran = True
            return arg + 1

        self.assertFalse(inspect.iscoroutinefunction(func))
        self.assertEqual(func(1), 2)
        self.assertTrue(ran)
        ran = False
        self.assertEqual(func(1), 2)
        self.assertFalse(ran)

    async def test_coro_func(self):
        ran = False

        @memoize
        async def afunc(arg):
            nonlocal ran
            ran = True
            return arg + 1

        self.assertTrue(inspect.iscoroutinefunction(afunc))
        self.assertEqual(await afunc(1), 2)
        self.assertTrue(ran)
        ran = False
        self.assertEqual(await afunc(1), 2)
        self.assertFalse(ran)



"""Base scales module for implementing scales."""

from concert.coroutines.base import background
from concert.devices.base import Device
from concert.base import Quantity, AccessorNotImplementedError
from concert.quantities import q


class WeightError(Exception):

    """Class for weighing errors."""

    pass


class Scales(Device):

    """Base scales class."""
    weight = Quantity(q.g, help="Weighted mass")

    def __init__(self):
        super(Scales, self).__init__()

    async def _get_weight(self):
        raise AccessorNotImplementedError


class TarableScales(Scales):

    """Scales which can be tared."""

    def __init__(self):
        super(TarableScales, self).__init__()

    @background
    async def tare(self):
        """
        tare()

        Tare the scales.
        """
        await self._tare()

    async def _tare(self):
        raise AccessorNotImplementedError



import tempfile
import shutil
import numpy as np
import os.path as op
from concert.coroutines.base import async_generate
from concert.storage import DummyWalker, DirectoryWalker, StorageError
from concert.tests import TestCase


class TestWalker(TestCase):

    def setUp(self):
        super(TestWalker, self).setUp()
        self.walker = DummyWalker()
        self.data = [0, 1]

    def check(self):
        truth = set([op.join('', 'foo', str(i)) for i in self.data])
        self.assertEqual(self.walker.paths, truth)

    async def test_coroutine(self):
        await self.walker.write(async_generate(self.data), dsetname='foo')
        self.check()


class TestDirectoryWalker(TestCase):

    def setUp(self):
        super(TestDirectoryWalker, self).setUp()
        self.path = tempfile.mkdtemp()
        self.walker = DirectoryWalker(root=self.path)
        self.data = np.ones((2, 2))

    def tearDown(self):
        shutil.rmtree(self.path)

    def test_directory_creation(self):
        self.walker.descend('foo')
        self.walker.descend('bar')
        self.assertTrue(op.exists(op.join(self.path, 'foo')))
        self.assertTrue(op.exists(op.join(self.path, 'foo', 'bar')))

    async def test_default_write(self):
        await self.walker.write(async_generate([self.data, self.data]))
        self.assertTrue(op.exists(op.join(self.path, 'frame_000000.tif')))
        self.assertTrue(op.exists(op.join(self.path, 'frame_000001.tif')))

        # Cannot write if directory is not empty
        with self.assertRaises(StorageError):
            await self.walker.write(async_generate([self.data]))

        # Make a new one ...
        self.walker.descend('foo')
        await self.walker.write(async_generate([self.data]))
        self.assertTrue(op.exists(op.join(self.path, 'foo', 'frame_000000.tif')))

    async def test_custom_write(self):
        await self.walker.write(async_generate([self.data]), dsetname='foo-{}.tif')
        self.assertTrue(op.exists(op.join(self.path, 'foo-0.tif')))

    def test_invalid_ascend(self):
        with self.assertRaises(StorageError):
            self.walker.ascend()

    async def test_dset_exists(self):
        await self.walker.write(async_generate([self.data]))
        with self.assertRaises(StorageError):
            await self.walker.write(async_generate([self.data]))

    async def test_same_directory_different_dset(self):
        await self.walker.write(async_generate([self.data]))
        await self.walker.write(async_generate([self.data]), dsetname='bar-{}.tif')

    async def test_dset_prefix(self):
        async def test_raises(dsetname):
            with self.assertRaises(ValueError):
                await self.walker.write(async_generate([self.data]), dsetname=dsetname)

        async def test_ok(dsetname):
            await self.walker.write(async_generate([self.data]), dsetname=dsetname)

        await test_ok('bar-{}.tif')
        await test_ok('baz-{:>06}.tif')
        await test_ok('spam-{0}')

        await test_raises('bar')
        await test_raises('bar-{')
        await test_raises('bar-}')
        await test_raises('bar-}{')
        await test_raises('bar-}{{}')



import asyncio
import copy
import logging
import threading
import time
import sys
import numpy as np

try:
    import gi
    gi.require_version('Ufo', '0.0')
    from gi.repository import Ufo
    import ufo.numpy
except ImportError as e:
    print(str(e))

try:
    from tofu.config import SECTIONS, GEN_RECO_PARAMS
    from tofu.util import get_reconstruction_regions
    from tofu.genreco import (CTGeometry, setup_graph, set_projection_filter_scale, make_runs,
                              DTYPE_CL_SIZE)
    from tofu.tasks import get_task
except ImportError:
    print("You must install tofu to use Ufo features, see 'https://github.com/ufo-kit/tofu.git'",
          file=sys.stderr)

from concert.base import Parameterizable, State, check, transition
from concert.config import PERFDEBUG
from concert.coroutines.base import background, async_generate, run_in_executor, run_in_loop, start
from concert.imageprocessing import filter_low_frequencies


LOG = logging.getLogger(__name__)


class PluginManager(object):

    """Plugin manager that initializes new tasks."""

    def __init__(self):
        self._wrapped = Ufo.PluginManager()

    def get_task(self, name, **kwargs):
        """
        Create a new task from plugin *name* and initialize with *kwargs*.
        """
        task = self._wrapped.get_task(name)
        task.set_properties(**kwargs)
        return task


class InjectProcess(object):

    """Process to inject NumPy data into a UFO processing graph.

    :class:`InjectProcess` can also be used as a context manager, in which
    case it will call :meth:`~.InjectProcess.start` on entering the manager
    and :meth:`~InjectProcess.wait` on exiting it.

    *graph* must either be a Ufo.TaskGraph or a Ufo.TaskNode object.  If it is
    a graph the input tasks will be connected to the roots, otherwise a new
    graph will be created. *scheduler* is one of the ufo schedulers, e.g.
    Ufo.Scheduler or Ufo.FixedScheduler.
    """

    def __init__(self, graph, get_output=False, output_dims=2, scheduler=None, copy_inputs=False):
        self.output_tasks = []
        self.sched = scheduler if scheduler else Ufo.Scheduler()
        self._started = False
        self.copy_inputs = copy_inputs

        if isinstance(graph, Ufo.TaskGraph):
            self.graph = graph
            roots = self.graph.get_roots()
        elif isinstance(graph, Ufo.TaskNode):
            self.graph = Ufo.TaskGraph()
            roots = [graph]
        else:
            msg = 'graph is neither Ufo.TaskGraph nor Ufo.TaskNode'
            raise ValueError(msg)

        # Initialize inputs
        self.input_tasks = {}
        self.ufo_buffers = {}
        for root in roots:
            self.input_tasks[root] = []
            self.ufo_buffers[root] = []
            num_inputs = root.get_num_inputs()
            for i in range(num_inputs):
                self.input_tasks[root].append(Ufo.InputTask())
                self.ufo_buffers[root].append(None)
                self.graph.connect_nodes_full(self.input_tasks[root][i], root, i)

        if get_output:
            for i, leave in enumerate(self.graph.get_leaves()):
                self.output_tasks.append(Ufo.OutputTask())
                self.output_tasks[-1].props.num_dims = output_dims
                self.graph.connect_nodes(leave, self.output_tasks[-1])

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.wait()

    async def __call__(self, producer):
        """Co-routine compatible consumer."""
        if not self._started:
            self.start()

        async for item in producer:
            await self.insert(item)
            yield await self.result(leave_index=0)

    def start(self, arch=None, gpu=None):
        """
        Run the processing in a new thread.

        Use :meth:`.push` to insert data into the processing chaing and
        :meth:`~InjectProcess.wait` to wait until processing has finished."""
        def run_scheduler(sched):
            sched.run(self.graph)

        if arch and gpu:
            sched = Ufo.FixedScheduler()
            sched.set_gpu_nodes(arch, [gpu])
        else:
            sched = self.sched

        self.thread = threading.Thread(target=run_scheduler, args=(sched,), daemon=True)
        self.thread.start()

        if not self._started:
            self._started = True

    @background
    async def insert(self, array, node=None, index=0):
        """
        Insert *array* into the *node*'s *index* input.

        .. note:: *array* must be a NumPy compatible array.
        """
        def _insert_real(array, node=None, index=0):
            if not node:
                if len(self.input_tasks) > 1:
                    raise ValueError('input_node cannot be None for graphs with more inputs')
                else:
                    node = list(self.input_tasks.keys())[0]
            if self.ufo_buffers[node][index] is None:
                # reverse shape from rows, cols to x, y
                self.ufo_buffers[node][index] = Ufo.Buffer.new_with_size(array.shape[::-1], None)
            else:
                self.ufo_buffers[node][index] = self.input_tasks[node][index].get_input_buffer()

            if array is not None:
                if self.copy_inputs:
                    array = np.copy(array, order='C')
                self.ufo_buffers[node][index].copy_host_array(array.__array_interface__['data'][0])
            self.input_tasks[node][index].release_input_buffer(self.ufo_buffers[node][index])

        await run_in_executor(_insert_real, array, node, index)

    @background
    async def result(self, leave_index=None):
        """Get result from *leave_index* if not None, all leaves if None. Returns a list of results
        in case *leave_index* is None or one result for the specified leave_index.
        """
        def _result_real(leave_index=None):
            if self.output_tasks:
                if leave_index is None:
                    indices = list(range(len(self.output_tasks)))
                else:
                    indices = [leave_index]
                results = []
                for index in indices:
                    buf = self.output_tasks[index].get_output_buffer()
                    if not buf:
                        return [None]
                    results.append(np.copy(ufo.numpy.asarray(buf)))
                    self.output_tasks[index].release_output_buffer(buf)

                if leave_index is not None:
                    results = results[0]

                return results

        return await run_in_executor(_result_real, leave_index)

    def stop(self):
        """Stop input tasks."""
        for input_tasks in list(self.input_tasks.values()):
            for input_task in input_tasks:
                input_task.stop()

    def wait(self):
        """Wait until processing has finished."""
        self.stop()
        self.thread.join()
        self._started = False


class FlatCorrect(InjectProcess):
    """
    Flat-field correction.
    """
    def __init__(self, dark, flat, absorptivity=True, fix_nan_and_inf=True, copy_inputs=False):
        self.dark = dark
        self.flat = flat
        self.ffc = get_task('flat-field-correct')
        self.ffc.props.fix_nan_and_inf = fix_nan_and_inf
        self.ffc.props.absorption_correct = absorptivity
        super().__init__(self.ffc, get_output=True, output_dims=2, copy_inputs=copy_inputs)

    async def __call__(self, producer):
        """Co-routine compatible consumer."""
        if self.dark is None or self.flat is None:
            raise InjectProcessError('dark and flat images must be set')

        if not self._started:
            self.start()

        first = True
        async for projection in producer:
            if projection.dtype != np.float32:
                projection = projection.astype(np.float32)
            await self.insert(projection, index=0)
            await self.insert(self.dark.astype(np.float32) if first else None, index=1)
            await self.insert(self.flat.astype(np.float32) if first else None, index=2)
            yield await self.result(leave_index=0)
            first = False


class GeneralBackprojectArgs(object):
    def __init__(self, center_position_x, center_position_z, number,
                 overall_angle=np.pi):
        self._slice_metric = None
        self._slice_metrics = ['min', 'max', 'sum', 'mean', 'var', 'std', 'skew',
                               'kurtosis', 'sag']
        self._z_parameters = SECTIONS['general-reconstruction']['z-parameter']['choices']
        for section in GEN_RECO_PARAMS:
            for arg in SECTIONS[section]:
                settings = SECTIONS[section][arg]
                default = settings['default']
                if default is not None and 'type' in settings:
                    default = settings['type'](default)
                setattr(self, arg.replace('-', '_'), default)
        self.y = 0
        self.height = None
        self.width = None
        self.center_position_x = center_position_x
        self.center_position_z = center_position_z
        self.number = number
        self.overall_angle = overall_angle

    @property
    def z_parameters(self):
        return self._z_parameters

    @property
    def slice_metrics(self):
        return self._slice_metrics

    @property
    def slice_metric(self):
        return self._slice_metric

    @slice_metric.setter
    def slice_metric(self, metric):
        if metric not in [None] + self.slice_metrics:
            raise GeneralBackprojectArgsError("Metric '{}' not known".format(metric))
        self._slice_metric = metric

    @property
    def z_parameter(self):
        return self._z_parameter

    @z_parameter.setter
    def z_parameter(self, name):
        if name not in self.z_parameters:
            raise GeneralBackprojectArgsError("Unknown z parameter '{}'".format(name))
        self._z_parameter = name


class GeneralBackproject(InjectProcess):

    """One backprojector worker.

    .. py:attribute:: args

        :class:`.GeneralBackprojectArgs` instance with arguments for reconstruction

    .. py:attribute:: resources

        `Ufo.Resources` instance

    .. py:attribute:: gpu_index

        use GPU with this index

    .. py:attribute:: do_normalization

        do flat correction of not

    .. py:attribute:: region

        User defined region which can override the one stored in `args`

    .. py:attribute:: copy_inputs

        if True copy images before they are inserted into UFO
    """

    def __init__(self, args, resources=None, gpu_index=0, do_normalization=False,
                 region=None, copy_inputs=False):
        if args.width is None or args.height is None:
            raise GeneralBackprojectError('width and height must be set in GeneralBackprojectArgs')
        scheduler = Ufo.FixedScheduler()
        if resources:
            scheduler.set_resources(resources)
        gpu = scheduler.get_resources().get_gpu_nodes()[gpu_index]

        self.args = copy.deepcopy(args)
        x_region, y_region, z_region = get_reconstruction_regions(self.args, store=True,
                                                                  dtype=float)
        set_projection_filter_scale(self.args)
        if region is not None:
            self.args.region = region
        LOG.debug('Creating reconstructor for gpu %d, region: %s', gpu_index, self.args.region)
        geometry = CTGeometry(self.args)
        if not self.args.disable_projection_crop:
            geometry.optimize_args()
        self.args = geometry.args

        regions = make_runs([gpu], [gpu_index], x_region, y_region, self.args.region,
                            DTYPE_CL_SIZE[self.args.store_type],
                            slices_per_device=self.args.slices_per_device,
                            slice_memory_coeff=self.args.slice_memory_coeff,
                            data_splitting_policy=self.args.data_splitting_policy)
        if len(regions) > 1:
            raise GeneralBackprojectError('Region does not fit to the GPU memory')

        graph = Ufo.TaskGraph()

        # Normalization
        self.ffc = None
        self.do_normalization = do_normalization
        if do_normalization:
            self.ffc = get_task('flat-field-correct', processing_node=gpu)
            self.ffc.props.fix_nan_and_inf = self.args.fix_nan_and_inf
            self.ffc.props.absorption_correct = self.args.absorptivity
            self._darks_averaged = False
            self._flats_averaged = False
            self.dark_avg = get_task('average', processing_node=gpu)
            self.flat_avg = get_task('average', processing_node=gpu)
            graph.connect_nodes_full(self.dark_avg, self.ffc, 1)
            graph.connect_nodes_full(self.flat_avg, self.ffc, 2)

        (first, last) = setup_graph(self.args, graph, x_region, y_region, self.args.region,
                                    source=self.ffc, gpu=gpu, index=gpu_index, do_output=False,
                                    make_reader=False)
        output_dims = 2
        if args.slice_metric:
            output_dims = 1
            metric = self.args.slice_metric
            if args.slice_metric == 'sag':
                metric = 'sum'
                gradient_task = get_task('gradient', processing_node=gpu, direction='both_abs')
                graph.connect_nodes(last, gradient_task)
                last = gradient_task
            measure_task = get_task('measure', processing_node=gpu, axis=-1, metric=metric)
            graph.connect_nodes(last, measure_task)
        elif first == last:
            # There are no other processing steps other than back projection
            LOG.debug('Only back projection, no other processing')
            graph = first

        super().__init__(graph, get_output=True, output_dims=output_dims, scheduler=scheduler,
                         copy_inputs=copy_inputs)

        if self.do_normalization:
            # Setup input tasks for normalization images averaging. Our parent picks up only the two
            # averagers and not the ffc's zero port for projections.
            self.input_tasks[self.ffc] = [Ufo.InputTask()]
            self.ufo_buffers[self.ffc] = [None]
            self.graph.connect_nodes_full(self.input_tasks[self.ffc][0], self.ffc, 0)

    async def _average(self, producer, node):
        what = 'darks' if node == self.dark_avg else 'flats'
        LOG.log(PERFDEBUG, 'Starting %s averaging', what)
        start = time.perf_counter()
        if not self._started:
            self.start()

        async for image in producer:
            await self.insert(image[self.args.y:self.args.y + self.args.height].astype(np.float32),
                              node=node, index=0)
        LOG.log(PERFDEBUG, '%s averaging duration: %g s', what, time.perf_counter() - start)

    @background
    async def average_darks(self, producer):
        await self._average(producer, self.dark_avg)
        self._darks_averaged = True

    @background
    async def average_flats(self, producer):
        await self._average(producer, self.flat_avg)
        self._flats_averaged = True

    async def __call__(self, producer):
        async def consume_volume():
            if not self._started:
                yield None
                return
            self.stop()
            st = time.perf_counter()
            for k in np.arange(*self.args.region):
                result = (await self.result())[0]
                if result is None:
                    LOG.warning('Not all slices received (last: %g)', k)
                    break
                yield result
            LOG.log(PERFDEBUG, 'Volume downloaded in: %.2f s', time.perf_counter() - st)
            self.wait()
            self._darks_averaged = False
            self._flats_averaged = False

        i = 0
        try:
            async for projection in producer:
                if i == 0:
                    st = time.perf_counter()
                    if not self._started:
                        self.start()
                    if self.do_normalization:
                        if not (self._darks_averaged and self._flats_averaged):
                            raise GeneralBackprojectError('Darks and flats not processed yet')
                        # Tell averagers to generate instead of consuming from now on
                        self.input_tasks[self.dark_avg][0].stop()
                        self.input_tasks[self.flat_avg][0].stop()
                i += 1

                projection = projection[self.args.y:self.args.y + self.args.height]
                if projection.dtype != np.float32:
                    projection = projection.astype(np.float32)
                # If ffc is None, node=None and that's OK because there is only one input which the
                # parent can deal with. Projection port of ffc task is zero, so no need for special
                # handling.
                await self.insert(projection, node=self.ffc, index=0)

                if i == self.args.number:
                    LOG.log(PERFDEBUG, 'Backprojected %d projections, duration: %.2f s', i,
                            time.perf_counter() - st)
                    async for item in consume_volume():
                        yield item
        except asyncio.CancelledError:
            # Stop scheduler to free up memory
            self.wait()
            raise


class GeneralBackprojectManager(Parameterizable):

    """Manage 3D recnstruction by back projection. The manager stores darks, flats and projections
    and spreads them to back projection workers in as many batches as needed in order not to
    overflow GPU memory.

    .. py:attribute:: args

        :class:`.GeneralBackprojectArgs` instance with arguments for reconstruction

    .. py:attribute:: average_normalization

       if False, use only one dark and flat image from their streams, otherwise average all of them

    .. py:attribute:: regions

        User defined regions (batches) for reconstruction, if None, batches are determined
        automatically

    .. py:attribute:: copy_inputs

        if True copy images before they are inserted into UFO
    """

    state = State(default='standby')

    def __init__(self, args, average_normalization=True, regions=None, copy_inputs=False):
        super().__init__()
        self.args = args
        self.regions = regions
        self.copy_inputs = copy_inputs
        self.projections = None
        self._resources = []
        self.volume = None
        self.average_normalization = average_normalization
        self.darks = []
        self.flats = []
        self._darks_condition = asyncio.Condition()
        self._flats_condition = asyncio.Condition()
        # Conditions for darks and flats need to signal that no more images will come and *done* is
        # how they do it
        self._darks_condition.done = False
        self._flats_condition.done = False
        self._num_received_projections = 0
        self._num_processed_projections = 0
        self._producer_condition = asyncio.Condition()
        self._processing_task = None
        self._regions = None

    @property
    def num_received_projections(self):
        """Number of received projections."""
        return self._num_received_projections

    @property
    def num_processed_projections(self):
        """Number of projections sent to backprojectors."""
        return self._num_processed_projections

    def _update(self):
        """Update the regions and volume sizes based on changed args or region."""
        st = time.perf_counter()
        x_region, y_region, z_region = get_reconstruction_regions(self.args)
        if not self._resources:
            self._resources = [Ufo.Resources()]
        gpus = np.array(self._resources[0].get_gpu_nodes())
        gpu_indices = np.array(self.args.gpus or list(range(len(gpus))))
        if min(gpu_indices) < 0 or max(gpu_indices) > len(gpus) - 1:
            raise ValueError('--gpus contains invalid indices')
        gpus = gpus[gpu_indices]
        if self.regions is None:
            self._regions = make_runs(gpus, gpu_indices, x_region, y_region, z_region,
                                      DTYPE_CL_SIZE[self.args.store_type],
                                      slices_per_device=self.args.slices_per_device,
                                      slice_memory_coeff=self.args.slice_memory_coeff,
                                      data_splitting_policy=self.args.data_splitting_policy,
                                      num_gpu_threads=self.args.num_gpu_threads)
        else:
            self._regions = self.regions
        offset = 0
        for batch in self._regions:
            for i, region in batch:
                if len(self._resources) < len(batch):
                    self._resources.append(Ufo.Resources())
                offset += len(np.arange(*region))
        if self.args.slice_metric:
            shape = (offset,)
        else:
            shape = (offset, len(np.arange(*y_region)), len(np.arange(*x_region)))
        if self.volume is None or shape != self.volume.shape:
            self.volume = np.empty(shape, dtype=np.float32)
        LOG.log(PERFDEBUG, 'Backprojector manager update duration: %g s', time.perf_counter() - st)

    async def _produce(self):
        """Produce projections for backprojectors."""
        for i in range(self.args.number):
            async with self._producer_condition:
                await self._producer_condition.wait_for(lambda: self._num_received_projections > i)
            yield self.projections[i]
            self._num_processed_projections = i + 1

    async def _consume(self, offset, producer):
        """Consume slices from individual backprojectors."""
        i = 0
        async for item in producer:
            self.volume[offset + i] = item
            i += 1

    def find_parameters(self, parameters, projections=None, metrics=('sag',), regions=None,
                        iterations=1, fwhm=0, minimize=(True,), z=None, method='powell',
                        method_options=None, guesses=None, bounds=None, store=True):
        """Find reconstruction parameters. *parameters* (see
        :attr:`.GeneralBackprojectArgs.z_parameters`) are the names of the parameters which should
        be found, *projections* are the input data and if not specified, the ones from last
        reconstruction are used. *z* specifies the height in which the parameter is looked for. If
        *store* is True, the found parameter values are stored in the reconstruction arguments.
        Optimization is done either brute-force if *regions* are not specified or one of the scipy
        minimization methods is used, see below.

        If *regions* are specified, they are reconstructed for the corresponding parameters and a
        metric from *metrics* list is applied. Thus, first parameter in *parameters* is
        reconstructed within the first region in *regions* and the first metric (see
        :attr:`.GeneralBackprojectArgs.slice_metrics`) in *metrics* is applied and so on. If
        *metrics* is of length 1 then it is applied to all parameters. *minimize* is a tuple
        specifying whether each parameter in the list should be minimized (True) or maximized
        (False). After every parameter is processed, the parameter optimization result is stored and
        the next parameter is optimized in such a way, that the result of the optimization of the
        previous parameter already takes place. *iterations* specifies how many times are all the
        parameters reconstructed. *fwhm* specifies the full width half maximum of the gaussian
        window used to filter out the low frequencies in the metric, which is useful when the region
        for a metric is large. If the *fwhm* is specified, the region must be at least 4 * fwhm
        large. If *fwhm* is 0 no filtering is done.

        If *regions* is not specified, :func:`scipy.minimize` is used to find the parameter, where
        the optimization method is given by the *method* parameter, *method_options* are passed as
        *options* to the minimize function and *guesses* are initial guesses in the order of the
        *parameters* list. If *bounds* are given, they represent the domains where to look for
        parameters, they are (min, max) tuples, also in the order of the *parameters* list. See
        documentation of :func:`scipy.minimize` for the list of minimization methods which support
        bounds specification. In this approach only the first in *metrics* is taken into account
        because the optimization happens on all parameters simultaneously, the same holds for
        *minimize*.
        """
        if projections is None:
            if self.projections is None:
                raise GeneralBackprojectManagerError('*projections* must be specified if no '
                                                     ' reconstructions have been done yet')
            projections = self.projections
        orig_args = self.args
        self.args = copy.deepcopy(self.args)

        if regions is None:
            # No region specified, do a real optimization on the parameters vector
            from scipy import optimize

            def score(vector):
                for (parameter, value) in zip(parameters, vector):
                    setattr(self.args, parameter.replace('-', '_'), [value])
                run_in_loop(self.backproject(async_generate(projections)))
                result = sgn * self.volume[0]
                LOG.info('Optimization vector: %s, result: %g', vector, result)

                return result

            self.args.z_parameter = 'z'
            z = z or 0
            self.args.region = [z, z + 1, 1.]
            self.args.slice_metric = metrics[0]
            sgn = 1 if minimize[0] else -1
            if guesses is None:
                guesses = []
                for parameter in parameters:
                    if parameter == 'center-position-x':
                        guesses.append(self.args.width / 2)
                    else:
                        guesses.append(0.)
            LOG.info('Guesses: %s', guesses)
            result = optimize.minimize(score, guesses, method=method, bounds=bounds,
                                       options=method_options)
            LOG.info('%s', result.message)
            result = result.x
        else:
            # Regions specified, reconstruct given regions for given parameters and simply search
            # for extrema of the given metrics
            self.args.z = z or 0
            if fwhm:
                for region in regions:
                    if len(np.arange(*region)) < 4 * fwhm:
                        raise ValueError('All regions must be at least 4 * fwhm large '
                                         'when fwhm is specified')
            result = []
            if len(metrics) == 1:
                metrics = metrics * len(parameters)
            if len(minimize) == 1:
                minimize = minimize * len(parameters)
            for i in range(iterations):
                for (parameter, region, metric, minim) in zip(parameters, regions,
                                                              metrics, minimize):
                    self.args.slice_metric = metric
                    self.args.z_parameter = parameter
                    self.args.region = region
                    run_in_loop(self.backproject(async_generate(projections)))
                    sgn = 1 if minim else -1
                    values = self.volume
                    if fwhm:
                        values = filter_low_frequencies(values, fwhm=fwhm)[2 * int(fwhm):
                                                                           -2 * int(fwhm)]
                    param_result = (np.argmin(sgn * values) + 2 * fwhm) * region[2] + region[0]
                    setattr(self.args, parameter.replace('-', '_'), [param_result])
                    if i == iterations - 1:
                        result.append(param_result)
                    LOG.info('Optimizing %s, region: %s, metric: %s, minimize: %s, result: %g',
                             parameter, region, metric, minim, param_result)

        LOG.info('Optimization result: %s', result)

        if store:
            for (parameter, value) in zip(parameters, result):
                setattr(orig_args, parameter.replace('-', '_'), [value])
        self.args = orig_args

        return result

    async def _copy_projections(self, producer):
        def copy_projection(projection):
            self.projections[self._num_received_projections] = projection

        async for projection in producer:
            if self._num_received_projections == 0:
                if not self.args.width:
                    (self.args.height, self.args.width) = projection.shape
                elif (self.args.height, self.args.width) != projection.shape:
                    raise GeneralBackprojectManagerError('Projections have different '
                                                         'shape from normalization images')
                in_shape = (self.args.number,) + projection.shape
                if (self.projections is None or in_shape != self.projections.shape
                        or projection.dtype != self.projections.dtype):
                    self.projections = np.empty(in_shape, dtype=projection.dtype)

            if self._num_received_projections < self.args.number:
                await run_in_executor(copy_projection, projection)
                self._num_received_projections += 1
                async with self._producer_condition:
                    self._producer_condition.notify_all()

    async def _distribute(self, reuse_normalization=False, do_normalization=False):
        """Distribute projections to multiple batches which may run on multiple GPUs. If
        *reuse_normalization* is True just feed the workers with the stored darks and flats, do not
        expect new streams. If *do_normalization* is True send darks and flats to backprojectors.
        """
        LOG.debug('Processing start')
        st = time.perf_counter()
        self._update()

        async def start_one(batch_index, region_index):
            """Start one backprojector with a specific GPU ID in a separate thread."""
            # first slice offset
            offset = 0
            for i in range(batch_index):
                for j, region in self._regions[i]:
                    offset += len(np.arange(*region))
            batch = self._regions[batch_index]
            offset += sum([len(np.arange(*reg)) for j, reg in batch[:region_index]])

            gpu_index, region = self._regions[batch_index][region_index]
            bp = GeneralBackproject(self.args,
                                    resources=self._resources[region_index],
                                    gpu_index=gpu_index,
                                    do_normalization=do_normalization,
                                    region=region,
                                    copy_inputs=self.copy_inputs)
            if do_normalization:
                if reuse_normalization:
                    darks = self.darks if self.average_normalization else self.darks[:1]
                    flats = self.flats if self.average_normalization else self.flats[:1]
                    darks_generator = async_generate(darks)
                    flats_generator = async_generate(flats)
                else:
                    darks_generator = self._produce_normalization(self.darks,
                                                                  self._darks_condition)
                    flats_generator = self._produce_normalization(self.flats,
                                                                  self._flats_condition)
                await asyncio.gather(bp.average_darks(darks_generator),
                                     bp.average_flats(flats_generator))
            await self._consume(offset, bp(self._produce()))

        LOG.debug('Reconstructing %d batches: %s', len(self._regions), self._regions)
        for batch_index in range(len(self._regions)):
            coros = []
            for region_index in range(len(self._regions[batch_index])):
                coros.append(start_one(batch_index, region_index))
            await asyncio.gather(*coros)

        # Process results
        duration = time.perf_counter() - st
        LOG.log(PERFDEBUG, 'Backprojectors duration: %.2f s', duration)
        in_size = self.projections.nbytes / 2 ** 20
        out_size = self.volume.nbytes / 2 ** 20
        LOG.log(PERFDEBUG, 'Input size: %g GB, output size: %g GB', in_size / 1024,
                out_size / 1024)
        LOG.log(PERFDEBUG, 'Performance: %.2f GUPS (In: %.2f MB/s, out: %.2f MB/s)',
                self.volume.size * self.projections.shape[0] * 1e-9 / duration,
                in_size / duration, out_size / duration)

    async def _produce_normalization(self, images, condition):
        i = 0
        while True:
            async with condition:
                # Either there is an image to consume, or last image has come and we have consumed
                # all of them
                await condition.wait_for(lambda: len(images) > i
                                         or condition.done and i == len(images))
                if condition.done and i == len(images):
                    break
            yield images[i]
            i += 1

    async def _copy_normalization(self, producer, images, condition):
        if not self._processing_task:
            # When we start a new backprojection job we need to clear both darks and flats because
            # we don't know which acquisition will be done first. This must be done here so that the
            # first image is not lost, which would be the case if we called it where the actual
            # _distribute is called.
            self.reset()
        try:
            async for image in producer:
                if not self.args.width:
                    (self.args.height, self.args.width) = image.shape
                if not self._processing_task:
                    # Start averaging before projection stream starts
                    self._processing_task = start(self._distribute(reuse_normalization=False,
                                                                   do_normalization=True))

                async with condition:
                    images.append(image)
                    condition.notify_all()

                if not self.average_normalization:
                    # We store only one
                    break

            # Last image has arrived, signal by *done* and notify waiting coros
            async with condition:
                condition.done = True
                condition.notify_all()
        except asyncio.CancelledError:
            if self._processing_task and not self._processing_task.cancelled():
                self._processing_task.cancel()
                self._processing_task = None
            raise

    def reset(self):
        """Reset state, clearing all pre-processing steps but keep projections and slices intact."""
        LOG.debug('Reset')
        if self._processing_task and not self._processing_task.done():
            raise GeneralBackprojectManagerError('Reconstruction running, cannot reset')
        del self.darks[:]
        del self.flats[:]
        self._darks_condition.done = False
        self._flats_condition.done = False
        self._processing_task = None
        self._num_received_projections = self._num_processed_projections = 0

    @background
    @check(source='standby', target='*')
    @transition(immediate='running', target='standby')
    async def update_darks(self, producer):
        """Get new darks from *producer*. Immediately start the reconstruction so that averaging
        starts.
        """
        await self._copy_normalization(producer, self.darks, self._darks_condition)

    @background
    @check(source='standby', target='*')
    @transition(immediate='running', target='standby')
    async def update_flats(self, producer):
        """Get new flats from *producer*. Immediately start the reconstruction so that averaging
        starts.
        """
        await self._copy_normalization(producer, self.flats, self._flats_condition)

    @background
    @check(source='standby', target='*')
    async def backproject(self, producer):
        """Backproject projections from *producer*."""
        self._state_value = 'running'
        self._num_received_projections = self._num_processed_projections = 0

        try:
            if not self._processing_task:
                self._processing_task = start(self._distribute(reuse_normalization=True,
                                                               do_normalization=self.darks
                                                               and self.flats))
            await asyncio.gather(self._copy_projections(producer), self._processing_task)
        except asyncio.CancelledError:
            if self._processing_task and not self._processing_task.cancelled():
                self._processing_task.cancel()
        finally:
            self._processing_task = None
            self._state_value = 'standby'


class InjectProcessError(Exception):

    """Errors related with :py:class:`.InjectProcess`."""


class GeneralBackprojectManagerError(Exception):
    pass


class GeneralBackprojectError(InjectProcessError):
    pass


class GeneralBackprojectArgsError(Exception):
    pass



"""
Remote add-ons working on image streams directly.
"""
import pickle
from concert.experiments.proxies.base import Proxy


class TangoProxy(Proxy):
    def __init__(self, tango_device):
        self._device = tango_device

    async def wait(self, count):
        await self._device.wait(count)


def ImageWriter(device):
    return device



"""
Cameras supported by the libuca library.
"""
import functools
import logging
import numpy as np
from concert.coroutines.base import background, run_in_executor
from concert.quantities import q
from concert.base import Parameter, Quantity, transition
from concert.helpers import Bunch
from concert.devices.cameras import base


LOG = logging.getLogger(__name__)


def _new_setter_wrapper(name, unit=None):
    async def _wrapper(instance, value):
        if await instance.get_state() == 'recording':
            raise base.CameraError('Changing parameters is not allowed while recording')

        if unit:
            value = value.to(unit)

        try:
            dic = {name: value.magnitude}
        except AttributeError:
            dic = {name: value}

        instance.uca.set_properties(**dic)

    return _wrapper


def _new_getter_wrapper(name, unit=None):
    async def _wrapper(instance):
        value = instance.uca.get_property(name)

        if unit:
            return q.Quantity(value, unit)

        return value

    return _wrapper


def _translate_gerror(func):
    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        from gi.repository import GLib
        try:
            return func(*args, **kwargs)
        except GLib.GError as ge:
            raise base.CameraError(str(ge))

    return _wrapper


class Camera(base.Camera):

    """libuca-based camera.

    All properties that are exported by the underlying camera are also visible.
    """

    def __init__(self, name, params=None):
        """
        Create a new libuca camera.

        The *name* is passed to the uca plugin manager.

        :raises CameraError: In case camera *name* does not exist.
        """

        super(Camera, self).__init__()

        import gi
        gi.require_version('Uca', '2.0')

        from gi.repository import GObject, GLib, Uca

        self._manager = Uca.PluginManager()

        params = params if params else {}

        try:
            self.uca = self._manager.get_camerah(name, params)
        except GLib.GError as ge:
            raise base.CameraError(str(ge))
        except Exception:
            raise base.CameraError("`{0}' is not a valid camera".format(name))

        units = {
            Uca.Unit.METER: q.m,
            Uca.Unit.SECOND: q.s,
            Uca.Unit.DEGREE_CELSIUS: q.celsius,
            Uca.Unit.COUNT: q.dimensionless,
            Uca.Unit.PIXEL: q.pixel,
        }

        parameters = {}

        for prop in self.uca.props:
            if prop.name in ('trigger-source', 'trigger-type', 'frames-per-second'):
                continue

            getter, setter, unit = None, None, None

            uca_unit = self.uca.get_unit(prop.name)

            if uca_unit in units:
                unit = units[uca_unit]

            if prop.flags & GObject.ParamFlags.READABLE:
                getter = _new_getter_wrapper(prop.name, unit)

            if prop.flags & GObject.ParamFlags.WRITABLE:
                setter = _new_setter_wrapper(prop.name, unit)

            name = prop.name.replace('-', '_')

            if uca_unit in units:
                parameters[name] = Quantity(unit, fget=getter, fset=setter, help=prop.blurb)
            else:
                parameters[name] = Parameter(fget=getter, fset=setter, help=prop.blurb)

        if parameters:
            self.install_parameters(parameters)

        class _Dummy(object):
            pass

        setattr(self.uca, 'enum_values', _Dummy())

        def get_enum_bunch(enum):
            enum_map = {}

            for key, value in list(enum.__enum_values__.items()):
                name = value.value_nick.upper().replace('-', '_')
                enum_map[name] = key

            return Bunch(enum_map)

        for prop in self.uca.props:
            if hasattr(prop, 'enum_class'):
                setattr(self.uca.enum_values, prop.name.replace('-', '_'),
                        get_enum_bunch(prop.default_value))

        self._uca_get_frame_rate = _new_getter_wrapper('frames-per-second')
        self._uca_set_frame_rate = _new_setter_wrapper('frames-per-second')

        # Invert the uca trigger source dict in order to return concert values
        trigger_dict = self.uca.enum_values.trigger_source.__dict__
        self._uca_to_concert_trigger = {v: k for k, v in list(trigger_dict.items())}
        self._uca_get_trigger = _new_getter_wrapper('trigger-source')
        self._uca_set_trigger = _new_setter_wrapper('trigger-source')

        self._record_shape = None
        self._record_dtype = None

    @background
    @transition(target='readout')
    async def start_readout(self):
        self.uca.start_readout()

    @background
    @transition(target='standby')
    async def stop_readout(self):
        self.uca.stop_readout()

    @background
    async def grab(self, index=None):
        return self.convert(await self._grab_real(index))

    def write(self, name, data):
        """Write NumPy array *data* for *name*."""
        raw = data.__array_interface__['data'][0]
        self.uca.write(name, raw, data.nbytes)

    async def _get_frame_rate(self):
        return await self._uca_get_frame_rate(self) / q.s

    async def _set_frame_rate(self, frame_rate):
        await self._uca_set_frame_rate(self, frame_rate * q.s)

    async def _get_trigger_source(self):
        uca_trigger = await self._uca_get_trigger(self)
        return self._uca_to_concert_trigger[uca_trigger]

    async def _set_trigger_source(self, source):
        uca_value = getattr(self.uca.enum_values.trigger_source, source)
        await self._uca_set_trigger(self, uca_value)

    @transition(target='recording')
    @_translate_gerror
    async def _record_real(self):
        self._record_shape = ((await self.get_roi_height()).magnitude,
                              (await self.get_roi_width()).magnitude)
        self._record_dtype = (np.uint16 if (await self.get_sensor_bitdepth()).magnitude > 8 else
                              np.uint8)
        self.uca.start_recording()

    @transition(target='standby')
    @_translate_gerror
    async def _stop_real(self):
        self.uca.stop_recording()

    @_translate_gerror
    async def _trigger_real(self):
        self.uca.trigger()

    @_translate_gerror
    async def _grab_real(self, index=None):
        array = np.empty(self._record_shape, dtype=self._record_dtype)
        data = array.__array_interface__['data'][0]

        if index is not None:
            if await run_in_executor(self.uca.readout, data, index):
                return array
            else:
                raise base.CameraError('No frame available')

        if await run_in_executor(self.uca.grab, data):
            return array
        else:
            raise base.CameraError('No frame available')



"""Sample exchanger is a device that will pick a sample from a sample holder and
transfer it onto another one.
"""
from concert.base import AccessorNotImplementedError
from concert.devices.base import Device


class SampleChanger(Device):

    """ A device that moves samples in and out from the sample holder."""

    def __init__(self):
        super(SampleChanger, self).__init__()

    async def _set_sample(self):
        raise AccessorNotImplementedError

    async def _get_sample(self):
        raise AccessorNotImplementedError



from concert.tests import TestCase
from concert.base import TransitionNotAllowed
from concert.devices.io.base import IODeviceError
from concert.devices.io.dummy import IO, Signal


class TestIO(TestCase):

    def setUp(self):
        self.io = IO(port_value=0)
        self.port = 0

    async def test_read(self):
        self.assertEqual(0, await self.io.read_port(self.port))

    async def test_write(self):
        value = 1
        await self.io.write_port(self.port, value)
        self.assertEqual(value, await self.io.read_port(self.port))

    async def test_non_existent_read(self):
        with self.assertRaises(IODeviceError):
            await self.io.read_port(1)

    async def test_non_existent_write(self):
        with self.assertRaises(IODeviceError):
            await self.io.write_port(1, 0)


class TestSignal(TestCase):

    def setUp(self):
        self.signal = Signal()

    async def test_on(self):
        await self.signal.on()
        self.assertEqual(await self.signal.get_state(), 'on')

        with self.assertRaises(TransitionNotAllowed):
            await self.signal.on()

    async def test_off(self):
        await self.signal.on()
        await self.signal.off()
        self.assertEqual(await self.signal.get_state(), 'off')

        with self.assertRaises(TransitionNotAllowed):
            await self.signal.off()

    async def test_trigger(self):
        await self.signal.trigger()
        self.assertEqual(await self.signal.get_state(), 'off')

        # Test forbidden state
        await self.signal.on()
        with self.assertRaises(TransitionNotAllowed):
            await self.signal.trigger()
